from __future__ import absolute_import
import copy
import traceback
import typing as tp  # NOQA
import warnings
import weakref

import numpy

import chainer
from chainer import _backprop
from chainer import backend
from chainer.backends import _cpu
from chainer.backends import cuda
from chainer.backends import intel64
from chainer import initializers
from chainer.initializers import constant
from chainer import types  # NOQA
import chainer.utils._collections
from chainer.utils import argument
import chainerx


def _check_grad_type(func, x, is_node_x, gx):
    # TODO(niboshi): Write comment about is_node_x
    assert gx is not None

    # TODO(kataoka): avoid `isinstance`
    x_data = None if isinstance(x, _ChainerxVariableNodeProps) else x.data

    # TODO(kataoka): Make _update_data_info store the array module.
    # ``is_node_x and x_data is None`` implies that the data array is not
    # retained.
    # ``not is_node_x and x_data is None`` implies that grad of uninitialized
    # variable is checked here.

    if x_data is None and not is_node_x:
        # TODO(kataoka): This should be an error.
        return
    if x.dtype is None or x.shape is None:
        # unretained Variable(None)
        # TODO(kataoka): This should be an error.
        return

    if not isinstance(gx, chainer.get_array_types()):
        msg = ('Type of grad is invalid:\n'
               + 'Expected: Any of {}\n'.format(chainer.get_array_types())
               + 'Actual: {}'.format(type(gx)))
        typ = TypeError
    elif x_data is not None and not chainer.is_arrays_compatible((gx, x_data)):
        msg = ('Type of data and grad mismatch\ngrad: %s != data: %s' %
               (type(gx), type(x_data)))
        typ = TypeError
    elif gx.dtype != x.dtype:
        msg = ('Dtype of data and grad mismatch\ngrad: %s != data: %s' %
               (gx.dtype, x.dtype))
        typ = TypeError
    elif gx.shape != x.shape:
        msg = ('Shape of data and grad mismatch\ngrad: %s != data: %s' %
               (gx.shape, x.shape))
        typ = ValueError
    else:
        return

    detail = ''
    if func:
        detail = 'Function `{0}` ({1}) has a bug.\n'.format(
            type(func)._impl_name, func.label)
        stack = func.stack
        if stack:
            detail += 'Stacktrace of the function is below:\n'
            for line in traceback.format_list(func.stack):
                detail += line
        detail += '''
Please report this error to the issue tracker with the stack trace,
the information of your environment, and your script:
https://github.com/chainer/chainer/issues/new.
'''

    raise typ(detail + msg)


def variable_repr(var):
    """Return the string representation of a variable.

    Args:
        var (~chainer.Variable): Input Variable.
    .. seealso:: numpy.array_repr
    """
    arr = _cpu._to_cpu(var.array)

    if var.name:
        prefix = 'variable ' + var.name
    else:
        prefix = 'variable'

    if arr is None:
        lst = 'None'
    elif arr.size > 0 or arr.shape == (0,):
        lst = numpy.array2string(arr, None, None, None, ', ', prefix + '(')
    else:  # show zero-length shape unless it is (0,)
        lst = '[], shape=%s' % (repr(arr.shape),)

    return '%s(%s)' % (prefix, lst)


def variable_str(var):
    """Return the string representation of a variable.

    Args:
        var (~chainer.Variable): Input Variable.
    .. seealso:: numpy.array_str
    """
    arr = _cpu._to_cpu(var.array)

    if var.name:
        prefix = 'variable ' + var.name
    else:
        prefix = 'variable'

    if arr is None:
        lst = 'None'
    else:
        lst = numpy.array2string(arr, None, None, None, ' ', prefix + '(')

    return '%s(%s)' % (prefix, lst)


class VariableNode(object):

    """Node in the backward computational graph representing a variable.

    This object represents a variable node in a computational graph. The node
    is used in error backpropagation (a.k.a. backprop) to determine which
    gradient to be passed to each function.

    A variable node is held by the corresponding :class:`~chainer.Variable`
    object, which is managed by users. :class:`~chainer.FunctionNode` objects
    that take the variable as an input also hold references to the variable
    node.

    Note that the node does not hold a reference to the corresponding data
    array in general. The data array is actually accessible by the node in the
    following cases.

    1. If there exists a :class:`~chainer.Variable` object that holds a
       reference to the variable node, the variable node holds a weak reference
       to the variable object, and thus the data array is accessible via the
       weak reference.
    2. If :meth:`retain_data` is called, the node holds a reference to the data
       array. It is mainly called by a function that needs the input or output
       data array in its backprop procedure.
       See :meth:`FunctionNode.retain_inputs()
       <chainer.FunctionNode.retain_inputs>`
       and :meth:`FunctionNode.retain_outputs()
       <chainer.FunctionNode.retain_outputs>` for more details.

    Users usually do not need to touch this variable node object. The
    computational graph is automatically managed by Chainer, and any interface
    that is beneficial for users is also provided by
    :class:`~chainer.Variable`.

    Args:
        variable (~chainer.Variable): The corresponding variable object.
        name (str): Name of the variable node.

    Attributes:
        dtype: Data type of the data array.
        shape: Shape of the data array.
        name (str): Name of the variable node.

    """

    _creator_node = None
    _data = None  # type: types.NdArray
    _rank = 0  # type: int
    # Name of the Function is assigned if this variable is a gradient generated
    # by an old-style Function
    _old_style_grad_generator = None  # type: str

    def __init__(self, variable, name, **kwargs):
        # type: (Variable, tp.Optional[str], **tp.Any) -> None

        if kwargs:
            argument.check_unexpected_kwargs(
                kwargs,
                grad='unexpected keyword argument "grad": '
                     'pass the gradient to Variable instead'
            )
        self._variable = weakref.ref(variable)
        self.name = name
        self._requires_grad = variable.requires_grad

        vdata = variable.data
        self._update_data_info(vdata)

    @property
    def creator(self):
        """Function object that created this variable node.

        When the function is implemented with the old-style API (i.e., it uses
        :class:`~chainer.Function` class),
        this property returns the :class:`~chainer.Function` object.
        The object is extracted from the :class:`~chainer.FunctionAdapter`
        object, so the returned object is not the function node, but instead
        the actual implementation of forward and backward procedures.

        When the function is implemented with the new-style API (i.e., it uses
        :class:`~chainer.FunctionNode` class),
        this property returns the function node
        object. In this case, the returned object is same as
        :attr:`creator_node`.

        .. warning::

           As of v3.0.0, when the creator is an old-style function, the
           following code is invalid:

           .. code-block:: python

              creator = v.creator
              v.creator = None
              ...
              v.creator = creator

           The point is that :class:`~chainer.FunctionNode` objects are used
           as nodes in the computational graph instead of
           :class:`~chainer.Function`, and each :class:`~chainer.Function`
           object only holds a *weak reference* to the corresponding
           :class:`~chainer.FunctionNode`.
           Since ``creator`` returns the :class:`~chainer.Function` object,
           the :class:`~chainer.FunctionNode` object is not kept by preserving
           ``creator``.

           The above code should be fixed as follows.

           .. code-block:: python

              creator_node = v.creator_node
              v.creator_node = None
              ...
              v.creator_node = creator_node

        """
        node = self._creator_node
        if node is None:
            return None

        if isinstance(node, chainer.function.FunctionAdapter):
            return node.function
        return node

    @creator.setter
    def creator(self, func):
        self.creator_node = func

    @property
    def creator_node(self):
        """Function node that has this variable as an output.

        See :class:`~chainer.FunctionNode` for the definition of a function
        node.

        """
        return self._creator_node

    @creator_node.setter
    def creator_node(self, func):
        if isinstance(func, chainer.Function):
            func = func.node
        self._creator_node = func
        if func is not None:
            self._rank = func.rank + 1

    @property
    def data(self):
        """Data array of the corresponding variable.

        If the data is not available, it returns ``None``.

        """
        return self._data

    @data.setter
    def data(self, d):
        self._data = d
        self._update_data_info(d)

    @property
    def grad(self):
        """Gradient array of the corresponding variable.

        If the variable is not available, it returns ``None``.

        """
        var = self._variable()
        return None if var is None else var.grad

    @property
    def grad_var(self):
        """Gradient variable of the corresponding variable.

        If the corresponding variable is not available, it return ``None``.

        """
        var = self._variable()
        return None if var is None else var.grad_var

    def _set_grad_var_if_available(self, g):
        var = self._variable()
        if var is not None:
            var._set_grad_var_without_check(g)

    @property
    def label(self):
        """Short text that represents the variable node."""
        if self.shape == ():
            return str(self.dtype)
        return '(%s), %s' % (', '.join(map(str, self.shape)),
                             str(self.dtype))

    @property
    def rank(self):
        return self._rank

    @property
    def requires_grad(self):
        """It indicates that ``grad`` will be set in backward calculation."""
        return self._requires_grad

    def get_variable(self):
        """Returns the corresponding :class:`~chainer.Variable` object.

        VariableNode object holds a weak reference of the variable object. If
        the reference is alive, it is returned by this property. Otherwise,
        this property creates a new :class:`~chainer.Variable` object from
        this node object and returns it.

        Returns:
            ~chainer.Variable: The variable object that refers this node.

        """
        var = self._variable()
        if var is not None:
            return var
        var = Variable._init_unchecked(
            self.data,
            name=self.name,
            requires_grad=self.requires_grad,
            node=self)
        return var

    def get_variable_or_none(self):
        """Returns the holding :class:`~chainer.Variable` object or ``None``.

        VariableNode object holds a weak reference of the variable object.If
        the reference is alive, it is returned by this property. Otherwise,
        returns ``None``.

        Returns:
            ~chainer.Variable: The variable object that refers this node.

        """
        return self._variable()

    def set_creator(self, creator):
        """Sets a :class:`~chainer.Function` object that created this node.

        This method is equivalent to ``self.creator = creator``. A
        :class:`~chainer.FunctionNode` object can also be passed.

        Args:
            creator (Function or FunctionNode): Function that has created this
                variable.

        """
        self.creator = creator

    def set_creator_node(self, creator_node):
        """Sets a :class:`~chainer.FunctionNode` object that created this node.

        This method is equivalent to ``self.creator_node = creator_node``. A
        :class:`~chainer.Function` object can also be passed, in which case the
        :attr:`Function.node <chainer.Function.node>` attribute is used.

        Args:
            creator_node (FunctionNode or Function): Function node that has
                this variable as an output.

        """
        self.creator_node = creator_node

    def unchain(self):
        """Deletes the reference to the creator of this variable node.

        This method is equivalent to ``self.creator_node = None``.

        """
        self.creator_node = None

    def retain_data(self):
        """Lets the node hold a reference to the underlying data array.

        This method gets the data array of the corresponding variable and keeps
        it. If the weak reference to the corresponding variable is dead, it
        raises an error.

        """
        variable = self._variable()
        if variable is not None:
            self.data = variable.data
        else:
            raise RuntimeError('cannot retain variable data: the variable has '
                               'been already released')

    def _update_data_info(self, d):
        if d is None:
            self.dtype = None
            self.shape = None
        else:
            self.dtype = d.dtype
            self.shape = d.shape

        # If the node has a reference to data, update it as well.
        if self._data is not None:
            self._data = d

    def _check_old_style_gradient(self):
        if self._old_style_grad_generator is not None:
            raise RuntimeError(
                'cannot twice-differentiate an old style Function "%s"' %
                self._old_style_grad_generator)


def _create_variable(data, name, grad, requires_grad, device):
    var = Variable(
        data, name=name, grad=grad, requires_grad=requires_grad)
    var.to_device(device)
    return var


class Variable(object):

    """__init__(data=None, *, name=None, grad=None, requires_grad=True)

    Array with a structure to keep track of computation.

    Every variable holds a data array of type either :class:`numpy.ndarray` or
    :class:`cupy.ndarray`.

    A variable object holds a data array and a
    :class:`~chainer.variable.VariableNode` object of
    a computational graph. If the variable is constructed by the user, the node
    is *root* and does not hold any parent. If the variable is constructed by a
    :class:`~chainer.FunctionNode` object (i.e., by calling functions under
    ``chainer.functions`` or user-defined functions), or by using operators
    (see the list below), the node holds a reference to its parent called
    :attr:`creator_node`.
    This reference is used in backpropagation to backtrack the graph.

    Users can disable (resp. enable) this chaining behavior by calling
    :func:`~chainer.no_backprop_mode` (resp.
    :func:`~chainer.force_backprop_mode`).
    In the former context, a variable never creates a computational graph,
    whereas in the latter context, it is forced to create.

    .. note::

        The following operators are defined for variable(s).

        * Indexing: ``a[slices]`` (:meth:`__getitem__`)
        * Addition: ``a + b`` (:meth:`__add__`, :meth:`__radd__`)
        * Subtraction: ``a - b`` (:meth:`__sub__`, :meth:`__rsub__`)
        * Multiplication: ``a * b`` (:meth:`__mul__`, :meth:`__rmul__`)
        * Division: ``a / b`` (:meth:`__div__`, :meth:`__rdiv__`, \
                               :meth:`__truediv__`, :meth:`__rtruediv__`)
        * Floor Division: ``a // b`` (:meth:`__floordiv__`, \
                                      :meth:`__rfloordiv__`)
        * Exponentiation: ``a ** b`` (:meth:`__pow__`, :meth:`__rpow__`)
        * Matrix Multiplication: ``a @ b`` (:meth:`__matmul__`, \
                                            :meth:`__rmatmul__`)
        * Negation (Arithmetic): ``- a`` (:meth:`__neg__`)
        * Absolute value: ``abs(a)`` (:meth:`__abs__`)

    Args:
        data (:ref:`ndarray`): Initial data array.
        name (str): Name of the variable.
        grad (:ref:`ndarray`): Initial gradient array.
        requires_grad (bool): Boolean indicating whether ``grad`` will be set
            in backward calculation.

    """

    # Cached value of `self.xp is chainerx`. It prevents from initializing
    # self._device as much as possible because it is really costly.
    _has_chainerx_array = False

    # Cached grad-stopped view of chainerx array. This is the return value
    # of `array` and `data` properties.
    _chainerx_nobp_array_cache = None

    # Cached grad-stopped view of the array returned by `grad` property.
    # It's a 2-element tuple, where the first is the original grad array and
    # the second is a grad-stopped view of the first. `grad` property returns
    # the second element.
    _chainerx_grad_cache = None

    _chainerx_name = None  # type: tp.Optional[str]

    # A NumPy, CuPy array cache to avoid redundant conversions between
    # NumPy/CuPy and ChainerX.
    # TODO(hvy): Avoid modifying this variable from outside this class.
    _chainerx_fallback_array = None

    # Used in non-ChainerX variables. The gradient array is stored in
    # this attribute on Variable.grad setter to delay creation of grad_var
    # instance.
    _grad = None

    def __init__(self, data=None, **kwargs):
        # type: (tp.Optional[types.NdArray], **tp.Any) -> None

        name, grad, requires_grad, grad_valid = argument.parse_kwargs(
            kwargs, ('name', None), ('grad', None), ('requires_grad', True),
            ('_grad_valid', True),
            volatile='volatile argument is not supported anymore. '
                     'Use chainer.using_config')
        # _grad_valid is for internal use, hence the prefix _.

        assert isinstance(requires_grad, bool)
        if data is not None:
            array_types = chainer.get_array_types()
            if not isinstance(data, array_types):
                msg = '{} or {} are expected. Actual: {}'.format(
                    ', '.join([str(at) for at in array_types[:-1]]),
                    array_types[-1], type(data))
                raise TypeError(msg)

        self._init_impl(
            data, None, name, grad, grad_valid, requires_grad, None, None)

    @staticmethod
    def _init_unchecked(
            data=None, device=None, name=None, grad=None, grad_valid=True,
            requires_grad=True, is_chainerx_array=None, node=None):
        """Creates a new :class:`Variable` without the validations for
        optimizing performance.
        """

        # Create a Variable without invoking __init__
        var = Variable.__new__(Variable)
        var._init_impl(
            data, device, name, grad, grad_valid, requires_grad,
            is_chainerx_array, node)
        return var

    def _init_impl(self, data, device, name, grad, grad_valid, requires_grad,
                   is_chainerx_array, node):
        # `device` must be of type chainer.backend.Device.
        # Check is skipped for performance.

        self._requires_grad = requires_grad  # type: bool
        self._loss_scale = None
        self._grad_var = None
        self._device = device
        # A flag to prevent grad from being used before calling cleargrad().
        # It becomes True when either
        # - cleargrad() is called, or
        # - zerograd() is called, or
        # - grad is set.
        # Note that it won't be True by merely initializing an uninitialized
        # Parameter.
        self._grad_valid = grad_valid

        if is_chainerx_array is None:
            is_chainerx_array = isinstance(data, chainerx.ndarray)

        if is_chainerx_array:
            if not requires_grad and grad is not None:
                raise ValueError(
                    'Cannot initialize a variable with gradients if the '
                    'require_grad argument is False.')
            self._set_chainerx_array(data, grad)  # type: ignore

            # ChainerX itself has own node objects, but not exposed to python.
            self._node = None  # type: tp.Optional[VariableNode]
            self._chainerx_name = name
        else:
            # Use a list as a data structure to hold the data array indirectly
            # to abstract its initialized/uninitialized state.
            self._data = [data]  # type: tp.List[tp.Optional[types.NdArray]]
            if node is None:
                self._node = VariableNode(self, name)
            else:
                self._node = node
            self._grad = grad

    def __copy__(self):
        return self._copy_to(Variable())

    def _copy_to(self, target):
        target.__dict__ = copy.copy(self.__dict__)
        target._node = VariableNode(target, self.name)
        return target

    def __reduce__(self):
        args = (
            self.array, self.name, self.grad, self._requires_grad, self.device)
        return _create_variable, args

    def __repr__(self):
        return variable_repr(self)

    def __str__(self):
        return variable_str(self)

    def _clear_chainerx(self):
        self._chainerx_nobp_array_cache = None
        self._chainerx_grad_cache = None
        self._chainerx_fallback_array = None

    def _ensure_grad_var_up_to_date(self):
        # For non-ChainerX, this method creates _grad_var if it's not yet
        # created and _grad is set.
        # For ChainerX, this method checks consistency between
        # _grad_var._data[0] and self._data[0].grad and recreates _grad_var
        # as necessary. (chainerx.ndarray.grad can be altered independently
        # from chainer)
        if self._has_chainerx_array:
            self._grad = None
            # Update gradient variable if it has not yet been initialized or
            # it happens to be dirty w.r.t. the actual gradient of the
            # underlying chainerx.ndarray.
            arr = self._data[0]
            actual_grad = (
                arr.grad
                if arr is not None and arr.is_grad_required()
                else None)
            if actual_grad is None:
                self._grad_var = None
            else:
                grad_var = self._grad_var
                old_grad = None if grad_var is None else grad_var._data[0]
                if actual_grad is not old_grad:
                    self._grad_var = Variable(
                        actual_grad,
                        requires_grad=actual_grad.is_backprop_required())
            return

        if self._grad_var is None:
            if self._grad is not None:
                self._grad_var = Variable(self._grad)

    def _set_chainerx_array(self, array, grad):
        # type: (tp.Optional[chainerx.ndarray], tp.Optional[chainerx.ndarray]) -> None # NOQA

        # Sets chainerx array and grad.
        assert array is None or isinstance(array, chainerx.ndarray)
        requires_grad = self._requires_grad

        self._grad = None

        if (not requires_grad
                and array is not None
                and array.is_backprop_required()):
            raise ValueError(
                'Cannot initialize a variable to not require '
                'gradients if the ChainerX array already requires '
                'backprop.')

        # Create a view of the given data to hold internally and modify.
        if array is None:
            self._data = [None]
        else:
            # If the array `array` is not connected to a graph, a view of it is
            # created and kept, in order not to change the no-graph status of
            # it. If the array is connected, the graph status is kept track of.
            if not array.is_backprop_required():
                array = array.view()
            if requires_grad:
                array.require_grad()
                if grad is not None:
                    array.set_grad(grad)
            self._data = [array]

        self._has_chainerx_array = True  # even if data is None
        self._chainerx_nobp_array_cache = None
        self._chainerx_grad_cache = None
        self._chainerx_fallback_array = None

    @property
    def device(self):
        """Device on which the data array of this variable reside."""
        # lazy initialization for performance
        if self._device is None:
            if self._data[0] is None:
                self._device = backend.CpuDevice()
            else:
                self._device = backend.get_device_from_array(self._data[0])
        return self._device

    @property
    def xp(self):
        # type: () -> tp.Optional[types.Xp]
        """Array module for the data array of this variable."""
        if self._has_chainerx_array:
            return chainerx
        else:
            device = self.device
            return None if device is None else device.xp

    @property
    def name(self):
        if self._has_chainerx_array:
            return self._chainerx_name
        return self._node.name

    @name.setter
    def name(self, n):
        if self._has_chainerx_array:
            self._chainerx_name = n
            return
        self._node.name = n

    def summary(self):
        if self.name:
            return '<variable %s>' % self.name
        else:
            return '<variable at 0x%x>' % id(self)

    def debug_print(self):
        """Display a summary of the stored data and location of the Variable"""

        msg = """{summary}
- device: {device}
- backend: {backend}
- shape: {shape}
- dtype: {dtype}
- statistics: {stats}
- grad: {grad}"""

        stats_msg = 'mean={0:.8f}, std={1:.8f}'

        array = self.array
        device = self.device
        with chainer.using_device(device):
            xp = device.xp

            if array is None:
                # `array` can be `None` if constructed without any arguments
                device = None
                backend = None
                stats = None
            else:
                device = getattr(array, 'device', 'CPU')
                backend = type(array)
                stats = stats_msg.format(float(xp.mean(array)),
                                         float(xp.std(array)))
            shape = getattr(array, 'shape', None)
            dtype = getattr(array, 'dtype', None)

            if self.grad is None:
                grad = None
            elif xp.all(self.grad == 0):
                grad = 0
            else:
                grad = stats_msg.format(float(xp.mean(self.grad)),
                                        float(xp.std(self.grad)))

        return msg.format(summary=self.summary(), device=device,
                          backend=backend, shape=shape, dtype=dtype,
                          stats=stats, grad=grad)

    def __pos__(self):
        return self

    def __len__(self):
        """Returns the first dimension of the data array.

        Returns:
            int: Number of the first dimension of the data array.

        """
        return len(self.array)

    @property
    def label(self):
        """Short text that represents the variable."""
        if self._has_chainerx_array:
            raise RuntimeError(
                'A variable of ChainerX does not provide a node label.')
        return self._node.label

    @property
    def creator(self):
        """Function implementation that created this variable.

        When this variable has been created by an old-style function (i.e., it
        is implemented as a subclass of :class:`Function`), this property
        returns that :class:`Function` object.

        When this variable has been created by a new-style function (i.e., it
        is implemented as a subclass of :class:`FunctionNode` class), this
        property returns that node object.

        """
        if self._has_chainerx_array:
            raise RuntimeError(
                'A variable of ChainerX does not provide a creator.')
        return self._node.creator

    @creator.setter
    def creator(self, func):
        if self._has_chainerx_array:
            raise RuntimeError(
                'A variable of ChainerX does not provide a creator.')
        self._node.creator = func

    @property
    def creator_node(self):
        """:class:`FunctionNode` object that created this variable.

        This property has a setter to which ``None`` can be set. Setting
        ``None`` to this property is equivalent to call :meth:`unchain`;
        it purges the variable from the function that created this variable.

        The setter also accepts the original :class:`FunctionNode` object that
        created this variable. For example, you can once set ``None`` to this
        property and then set the original value again.

        .. note::
           Setting an irrelevant :meth:`FunctionNode` object does not emit any
           error immediately, whereas the behavior is undefined. Do not set
           a :meth:`FunctionNode` object that did not create this variable
           object.

        """
        if self._has_chainerx_array:
            raise RuntimeError(
                'A variable of ChainerX does not provide a creator_node.')
        return self._node._creator_node

    @creator_node.setter
    def creator_node(self, func):
        if self._has_chainerx_array:
            raise RuntimeError(
                'A variable of ChainerX does not provide a creator_node.')
        self._node.creator_node = func

    @property
    def array(self):
        # type: () -> tp.Optional[types.NdArray]
        """The underlying data array.

        It is either :class:`numpy.ndarray` or :class:`cupy.ndarray` object,
        or ``None`` if the variable in in an uninitialized state.

        """
        # For ChainerX, this property always returns a grad-stopped view.
        # The view is cached to reduce potential overhead.
        if self._has_chainerx_array:
            if (self._chainerx_nobp_array_cache is None
                    and self._data[0] is not None):
                self._chainerx_nobp_array_cache = (
                    self._data[0].as_grad_stopped())  # type: ignore
            return self._chainerx_nobp_array_cache

        return self._data[0]

    @array.setter
    def array(self, d):
        # type: (tp.Optional[types.NdArray]) -> None

        if self._has_chainerx_array:
            d_old = self._data[0]
            if (d_old is not None
                    and (d_old.is_backprop_required()  # type: ignore
                         or d.is_backprop_required())):  # type: ignore
                raise ValueError(
                    'Cannot update the array of a Variable if either the '
                    'existing or the new array requires backprop.')

            self._set_chainerx_array(d, None)  # type: ignore
        else:
            self._node._update_data_info(d)  # type: ignore # _node doesn't have value when xp is chainerx # NOQA
            self._data[0] = d
            self._has_chainerx_array = False

    @property
    def chx_array(self):
        """A view of the raw ChainerX array.

        In contrary to :data:`Variable.array` which is always disconnected,
        the array represented by this attribute may be connected to the
        computational graph.

        It is a view, so it has a distinct gradient from the original array.

        If this attribute is queried on a :class:`Variable` with a non-ChainerX
        array, :class:`ValueError` will be raised.
        """
        if not self._has_chainerx_array:
            raise ValueError(
                'chx_array is not available for Variable with '
                'non-ChainerX array.')
        return self._data[0].view()

    @property
    def data(self):
        # type: () -> tp.Optional[types.NdArray]
        """The underlying data array (equivalent to :attr:`array`).

        Note that using this attribute directly is discouraged; use
        :attr:`array` instead. Using :attr:`array`, you can find an error
        earlier when your code mixes up Variable and ndarray because
        ndarray does not have an attribute ``.array`` while it has
        ``.data``.

        """
        return self.array

    @data.setter
    def data(self, d):
        # type: (types.NdArray) -> None

        self.array = d

    def _set_chainerx_grad(self, g, from_grad_var):
        # Assigns chainerx.ndarray.grad.
        #
        # If the main array is connected to the graph, in order to enable
        # double-backprop, the grad will also be backprop-required
        # (a view is created not to affect the given grad).
        # If the given grad is from a grad_var, this operation is skipped,
        # as the status of the given grad reflects the necessity of
        # double-backprop.
        assert self.xp is chainerx
        if not self._requires_grad and g is not None:
            raise RuntimeError(
                'Cannot set the gradient of a variable that is flagged to not '
                'require one.')
        arr = self._data[0]
        if arr is None:
            if g is not None:
                raise RuntimeError(
                    'Cannot set a gradient to an empty variable')
        elif arr.is_backprop_required():
            # If g is grad-stopped, require grad on it.
            # Make a view in order not to affect the input.
            if (g is not None
                    and not from_grad_var
                    and not g.is_backprop_required()):
                g = g.view().require_grad()
            arr.set_grad(g)

    def _set_grad_without_check(self, g):
        if self._has_chainerx_array:
            self._set_chainerx_grad(g, False)
            self._grad_var = None
            self._grad_valid = True
            return

        self._grad = g
        self._grad_var = None
        self._grad_valid = True

    @property
    def grad(self):
        # type: () -> tp.Optional[types.NdArray]
        """Gradient array of this variable.

        Note that this property returns the underlying array of the gradient
        variable instead of the gradient variable itself; to get/set
        gradient variable, use :attr:`grad_var` instead.

        If the underlying array is a :class:`chainerx.ndarray` and
        requires_grad is false, trying to access the gradient will results in
        and error.

        """
        if not self._grad_valid:
            raise RuntimeError(
                'Cannot retrieve Variable.grad. '
                'Either it must be set manually or Variable.cleargrad() '
                'must be called beforehand.')

        if self._has_chainerx_array:
            arr = self._data[0]
            if arr is None or not arr.is_backprop_required():
                self._chainerx_grad_cache = None
                return None

            actual_grad = arr.grad

            if actual_grad is None:
                self._chainerx_grad_cache = None
                return None

            # If grad is cached and the actual grad has not changed, return
            # the cache.
            if self._chainerx_grad_cache is not None:
                orig_grad, grad_stopped_grad = self._chainerx_grad_cache
                if orig_grad is actual_grad:
                    return grad_stopped_grad

            # Update the cache
            grad_stopped_grad = actual_grad.as_grad_stopped()
            self._chainerx_grad_cache = (actual_grad, grad_stopped_grad)

            return grad_stopped_grad

        if self._grad_var is not None:
            return self._grad_var.array
        return self._grad

    @grad.setter
    def grad(self, g):
        # type: (tp.Optional[types.NdArray]) -> None
        if g is not None:
            _check_grad_type(None, self, False, g)
        self._set_grad_without_check(g)

    def _set_grad_var_without_check(self, gv):
        if self._has_chainerx_array:
            self._set_chainerx_grad(
                None if gv is None else gv._data[0],
                True)
            self._grad_var = gv
            return

        self._grad_var = gv
        self._grad = None if gv is None else gv.array

    @property
    def grad_var(self):
        # type: () -> tp.Optional["Variable"]
        """Gradient variable."""
        self._ensure_grad_var_up_to_date()
        return self._grad_var

    @grad_var.setter
    def grad_var(self, g):
        # type: (tp.Optional["Variable"]) -> None
        if g is not None:
            _check_grad_type(None, self, False, g.array)
        self._set_grad_var_without_check(g)

    @property
    def shape(self):
        return self.array.shape

    @property
    def ndim(self):
        return self.array.ndim

    @property
    def size(self):
        return self.array.size

    @property
    def dtype(self):
        return self.array.dtype

    @property
    def rank(self):
        if self._has_chainerx_array:
            raise RuntimeError(
                'A variable of ChainerX does not provide a node rank.')
        return self._node.rank

    @property
    def node(self):
        if self._has_chainerx_array:
            raise RuntimeError(
                'A variable of ChainerX does not provide a node.')
        return self._node

    @property
    def requires_grad(self):
        """It indicates that ``grad`` will be set in backward calculation."""
        return self._requires_grad

    @property
    def T(self):
        """Transposition of this variable."""
        return chainer.functions.transpose(self)

    def to_cpu(self):
        """Copies the data and gradient arrays to CPU."""
        self.to_device(backend.CpuDevice())

    def to_gpu(self, device=None):
        """Copies the data and gradient arrays to specified GPU.

        Args:
            device: Target device specifier. If omitted, the current device is
                used.

        """
        cuda.check_cuda_available()
        self.to_device(cuda._get_device_or_current(device))

    def to_intel64(self):
        """Copies the data and gradient arrays to intel64 specific mdarray.

        If the array is not suited for intel64, it will be converted to
        :class:`numpy.ndarray`.
        """
        intel64.check_ideep_available()
        self.to_device(intel64.Intel64Device())

    def to_chx(self):
        """Converts the array and gradient to ChainerX arrays without copy.

        This method converts the underlying array and gradient to
        :class:`chainerx.ndarray` on the same physical device. It does nothing
        if the array held by the Variable object is already a ChainerX array.
        The new array is a view of the original one.

        """
        self._to_chx(allow_unchaining=False)

    def _to_chx(self, allow_unchaining):
        if not chainerx.is_available():
            raise RuntimeError('ChainerX is not available.')

        if self._has_chainerx_array:
            return

        if not allow_unchaining and self.creator is not None:
            raise RuntimeError(
                'A variable with a creator cannot be converted into ChainerX '
                'array')

        self._to_device(
            backend.ChainerxDevice.from_fallback_device(self.device),
            allow_unchaining)

    def from_chx(self):
        """Converts the array and gradient to non-ChainerX arrays without copy.

        This method converts the underlying ChainerX array and gradient
        residing in either a ``native`` or ``cuda`` device to NumPy or CuPy
        arrays respectively, on their same physical device. It does nothing
        if the array held by the Variable object is not a ChainerX array. The
        new array is a view of the original one.

        Raises an error if such a conversion is not supported for the device.

        """
        self._from_chx(allow_unchaining=False)

    def _from_chx(self, allow_unchaining):
        if not self._has_chainerx_array:
            return

        if not allow_unchaining and self._data[0].is_backprop_required():
            raise RuntimeError(
                'Cannot convert from a Variable with a ChainerX array that is '
                'connected to a graph.')

        self.to_device(self.device.fallback_device)

    def to_device(self, device):
        """Copies the data and gradient arrays to specified device.

        Args:
            device: Target device specifier. See
                :func:`~chainer.get_device` for available values.

        """
        self._to_device(device, allow_unchaining=False)

    def _to_device(self, device, allow_unchaining):
        device = chainer.get_device(device)

        was_chainerx = self._has_chainerx_array
        is_chainerx = device.xp is chainerx

        if not allow_unchaining:
            if was_chainerx and not is_chainerx:
                chx_arr = self._data[0]
                if chx_arr is not None and chx_arr.is_backprop_required():
                    raise RuntimeError(
                        'A variable of a ChainerX array which requires '
                        'gradients cannot be copied into non-chainerx device '
                        '({}).'.format(device))
            elif not was_chainerx and is_chainerx:
                arr = self._data[0]
                if arr is not None and self.creator is not None:
                    raise RuntimeError(
                        'A variable of a non-ChainerX array which is '
                        'connected to a graph cannot be copied to a ChainerX '
                        'device ({}).'.format(device))

        arr = self._data[0]
        grad_var = self.grad_var

        if was_chainerx and not is_chainerx:
            self._clear_chainerx()
            self._node = VariableNode(self, self._chainerx_name)
        elif not was_chainerx and is_chainerx:
            self._chainerx_name = self._node.name

        self._device = device
        self._has_chainerx_array = is_chainerx

        if arr is None:
            return

        if backend.get_device_from_array(arr) == device:
            return

        new_arr = device.send(arr)
        if is_chainerx:
            if grad_var is None:
                new_grad = None
            else:
                new_grad = device.send(grad_var._data[0])
            self._set_chainerx_array(new_arr, new_grad)
        else:
            self._data = [new_arr]
            if grad_var is not None:
                grad_var._to_device(device, allow_unchaining=allow_unchaining)
                # _grad has been invalidated by the line above.
                self._grad = grad_var.array

        # ensure that the node tracks the device migration
        node = self._node
        if is_chainerx:
            # ChainerX itself has own node objects,
            # ensure that the node is disconnected with this variable.
            if node is not None:
                # Disconnect by replacing with an alternative of dead weakref
                node._variable = lambda: None
                self._node = None
        else:
            if node._data is not None:
                node.retain_data()

    def cleargrad(self):
        """Clears the gradient array."""
        self.grad_var = None
        self._grad_valid = True

    def zerograd(self):
        """Initializes the gradient array by zeros.


        Note that the gradient variable is unchained from the computational
        graph by this method, because this operation breaks the backprop
        validity.

        .. deprecated:: v1.15
           Use more efficient  :meth:`cleargrads` instead.

        """
        warnings.warn(
            'Variable.zerograd is deprecated. Use Variable.cleargrad instead.',
            DeprecationWarning)

        arr = self.array
        if arr is None:
            self._grad_valid = True
            return

        if self._has_chainerx_array:
            gv = self.grad_var
            if gv is None:
                self.grad = chainerx.zeros_like(
                    arr, device=self.device.device)
            else:
                gv._data[0].fill(0)
        else:
            with chainer.using_device(self.device):
                xp = self.device.xp
                if self._grad is None:
                    self._grad = xp.zeros_like(arr)
                    self._grad_var = None
                else:
                    gv = self._grad_var
                    if gv is not None:
                        gv.unchain()
                    self._grad.fill(0)
        self._grad_valid = True

    def copydata(self, var):
        """Copies the data array from given source variable.

        This method copies the data array from given variable to this variable.
        The copy is done even if the arrays reside on different devices,
        including across the host and a GPU device. If this variable has an
        uninitialized data array, this method initializes it by the data array
        of the given variable. Similarly, if the given variable has an
        uninitialized data array, this method initializes it by the data array
        of this variable (``self``). If both are uninitialized, this method
        does nothing.

        Args:
            var (~chainer.Variable): Source variable.

        """
        src = var.array
        dst = self.array
        if src is None:
            if dst is None:
                return
            var.initialize(self.shape)
            src = var.array
        elif dst is None:
            self.initialize(src.shape)
            dst = self.array
        backend.copyto(dst, src)

    def addgrad(self, var):
        """Accumulates the gradient array from given source variable.

        This method adds the gradient of a given variable to the gradient of
        this variable. The accumulation is even done across the host and
        different devices. If this variable has uninitialized data/grad arrays,
        this method initializes it with the shape of the given variable and
        then accumulates the gradient.

        Args:
            var (~chainer.Variable): Source variable.

        """
        dst_device = self.device
        is_chainerx = dst_device.xp is chainerx

        if is_chainerx != (var.device.xp is chainerx):
            raise RuntimeError(
                'Variable.addgrad does not support addition between '
                'gradients on non-ChainerX and ChainerX devices.\n'
                'Adding gradient to: {}\n'
                'Adding gradient from: {}'.format(
                    dst_device, var.device))

        if var.grad is None:
            return

        src = var.grad_var

        if self.array is None:
            self.initialize(var.shape)

        dst = self.grad_var
        src_device = src.device
        if src_device != dst_device:
            src = chainer.functions.copy(src, dst_device)
        self.grad_var = src if dst is None else src + dst

    def set_creator(self, gen_func):
        """Notifies the variable that the given function is its creator.

        Args:
            gen_func (Function): Function object that creates this variable as
                one of its outputs.

        """
        if self._has_chainerx_array:
            raise RuntimeError(
                'A variable of ChainerX does not provide a creator.')
        self._node.set_creator(gen_func)

    def set_creator_node(self, fnode):
        """Notifies the variable that the given node is its creator.

        Args:
            fnode (FunctionNode): Function node that has this variable as an
                output.

        """
        if self._has_chainerx_array:
            raise RuntimeError(
                'A variable of ChainerX does not provide a creator node.')
        self._node.set_creator_node(fnode)

    def backward(self, retain_grad=False, enable_double_backprop=False,
                 loss_scale=None):
        """Runs error backpropagation (a.k.a.\\  backprop) from this variable.

        On backprop,
        :meth:`FunctionNode.backward() <chainer.FunctionNode.backward>`
        is called on each :class:`~chainer.FunctionNode` object appearing in
        the backward graph starting from this variable.
        The backward graph is represented by backward
        references from variable nodes to their creators, and from function
        nodes to their input variable nodes. The backprop stops at all root
        nodes. Some function nodes set ``None`` as gradients of some inputs,
        where further backprop does not take place at such inputs.

        This method uses :data:`grad` as the initial error array. User can
        manually set a gradient array before calling this method.
        If the shape of :data:`data` is ``()`` (i.e., it is scalar) and
        :data:`grad` is ``None``, then this method automatically complements
        1.0 as the initial error. This is useful on starting backprop from
        some scalar loss value.

        From v3, this method supports *differentiable backprop* (a.k.a. double
        backprop, grad of grads). To enable it, pass
        ``enable_double_backprop=True``.

        Args:
            retain_grad (bool): If ``True``, the gradient arrays of all
                intermediate variables are kept.
                Otherwise, :data:`~chainer.Variable.grad` of the
                intermediate variables are set to ``None`` on appropriate
                timing, which may reduce the maximum memory consumption.

                In most cases of training some models, the purpose of backprop
                is to compute gradients of parameters, not of all variables,
                and therefore it is recommended that this flag be set to
                ``False``.
            enable_double_backprop (bool): *(Added in v3.0)* If ``True``,
                computational trace of the whole backpropagation procedure is
                recorded to the computational graph so that one can further do
                backpropagation from the resulting gradients. Note that
                enabling it results in larger memory consumption needed to
                store the gradients w.r.t intermediate variables that are
                required for the second gradient computation.
            loss_scale (float): Loss scaling factor. Loss scaling is a usefull
                technique to mitigate vanishing gradient issue that tends to
                happen when low precision data type like float16 is used during
                training. If you set loss scaling factor, gradients of loss
                values are to be multiplied by the factor before backprop
                starts. The factor is propagated to whole gradients in a
                computational graph along the backprop. The gradients of
                parameters are divided by the factor just before the parameters
                are to be updated.
        """
        if self._has_chainerx_array:
            if retain_grad:
                raise RuntimeError(
                    'retain_grad is not supported for ChainerX array.')
            arr = self._data[0]
            assert isinstance(arr, chainerx.ndarray)
            # pybind has issues when converting int -> opt<float>
            if loss_scale:
                loss_scale = float(loss_scale)
            chainerx.backward(
                arr, enable_double_backprop=enable_double_backprop,
                loss_scale=loss_scale)
            return

        # Initialize error by 1, if this is a loss variable
        if self.array.size == 1 and self.grad_var is None:
            if self.array.ndim != 0:
                warnings.warn(
                    'Treating a variable with only one element as a scalar'
                    ' in Variable.backward is deprecated. A scalar variable'
                    ' must be a 0-dimensional array. Apply'
                    ' chainer.functions.squeeze to obtain a scalar variable.'
                    ' If the size of this variable accidentally becomes one,'
                    ' set zero to grad.',
                    DeprecationWarning)
            with chainer.using_device(self.device):
                self.grad = self.device.xp.ones_like(self.array)
            if loss_scale is not None:
                self.grad *= loss_scale

        node = self.node
        grad_var = self.grad_var
        self.grad_var = None

        with chainer.using_config('enable_backprop', enable_double_backprop):
            # TODO(kataoka): The following line should not pass grad_var = None
            # to _backprop_to_all, but it is working because grad_var is
            # immediately popped away as None = _backprop_utils._reduce([None])
            _backprop._backprop_to_all(
                [(node, grad_var)], retain_grad, loss_scale)

    def item(self):
        """Converts the variable with one element to a Python scalar.

        This will incur host-device synchronization.

        Returns:
            int or float: The element of the array.

        """
        return self.array.item()

    def mean(self, axis=None, weights=None, keepdims=False):
        """Calculate weighted average of array elements over a given axis.

        .. seealso::
           :func:`chainer.functions.average` for full documentation,

        """
        return chainer.functions.average(self, axis, weights, keepdims)

    def reshape(self, *shape):
        """Returns a variable of a different shape and the same content.

        .. seealso::
           :func:`chainer.functions.reshape` for full documentation,

        """
        if len(shape) == 1 and isinstance(shape[0], (tuple, list)):
            shape = shape[0]
        return chainer.functions.reshape(self, shape)

    def transpose(self, *axes):
        """Permute the dimensions of an input variable without copy.

        .. seealso::
           :func:`chainer.functions.transpose` for full documentation.

        """
        if len(axes) == 0:
            axes = None
        elif len(axes) == 1 and (isinstance(axes[0], (tuple, list)) or
                                 axes[0] is None):
            axes = axes[0]
        return chainer.functions.transpose(self, axes)

    def unchain(self):
        """Deletes the reference to the creator of this variable.

        This method deletes the reference to the creator from the corresponding
        variable node. Unlike :meth:`unchain_backward`, it does not backtrack
        the graph.

        This method is equivalent to ``self.creator_node = None``.

        """
        if self._has_chainerx_array:
            raise RuntimeError(
                'A variable of ChainerX does not provide an unchain method.')
        self.creator_node = None

    def unchain_backward(self):
        """Deletes references between variable nodes and functions backward.

        After this method completes, intermediate variable nodes and functions
        that are not referenced from anywhere are deallocated by reference
        count GC. Also this variable itself deletes the reference to its
        creator function from the node, i.e. the node becomes root in the
        computation graph. It indicates that backprop after unchaining stops at
        this variable. This behavior is useful to implement truncated BPTT.

        """
        if self._has_chainerx_array:
            raise RuntimeError(
                'A variable of ChainerX does not provide an unchain_backward '
                'method.')
        cand_funcs = []
        seen_set = set()

        def add_cand(cand):
            if cand is not None and cand not in seen_set:
                cand_funcs.append(cand)
                seen_set.add(cand)

        add_cand(self.creator_node)

        while cand_funcs:
            func = cand_funcs.pop()
            for var in func.inputs:
                add_cand(var.creator_node)
            func.unchain()

    def retain_data(self):
        """Lets the corresponding variable node keep the underlying array."""
        if self._has_chainerx_array:
            raise RuntimeError(
                'A variable of ChainerX does not provide a retain_data '
                'method.')
        self._node.data = self._data[0]

    def _error_nobp_op(self, op):
        raise TypeError(
            'Variables do not support {} operator. '
            'You could use `array` attribute instead.'.format(op))

    def __lt__(self, other):
        """This operator is not supported in Variables."""
        self._error_nobp_op('<')

    def __le__(self, other):
        """This operator is not supported in Variables."""
        self._error_nobp_op('<=')

    def __eq__(self, other):
        """This operator is not supported in Variables."""
        self._error_nobp_op('==')

    def __ne__(self, other):
        """This operator is not supported in Variables."""
        self._error_nobp_op('!=')

    def __gt__(self, other):
        """This operator is not supported in Variables."""
        self._error_nobp_op('>')

    def __ge__(self, other):
        """This operator is not supported in Variables."""
        self._error_nobp_op('>=')

    def __nonzero__(self):
        """This operator is not supported in Variables."""
        # Python 2.x
        raise TypeError(
            'Variables cannot be evaluated as Python bool.')

    def __bool__(self):
        """This operator is not supported in Variables."""
        # Python 3.x
        raise TypeError(
            'Variables cannot be evaluated as Python bool.')

    __array_priority__ = 200  # type: int
    __hash__ = None  # type: tp.Callable[[object], int]


class Parameter(Variable):

    """Parameter variable that can be registered to a link.

    Parameter is a subclass of :class:`Variable`. It almost behaves as same
    as a usual variable except that a parameter can be registered to a
    :class:`~chainer.Link` object just by assigning it to an attribute of
    the link within an :meth:`~chainer.Link.init_scope` context.

    Parameter also supports an initialization by an initializer. It can have
    two initializers: one for the data array, and the other for the gradient
    array. The initializer only specifies the way of filling the elements of
    these arrays, and the shape information is specified at the initialization
    point.

    When a link that the parameter has been registered to is passed to an
    :class:`~chainer.GradientMethod`, an update rule is set to the parameter.
    This update rule specifies how to update the data array of the parameter
    using its gradient array.

    Args:
        initializer (~chainer.Initializer or :ref:`ndarray`):
            Initializer of the data array. If ``shape`` is given, this
            initializer is immediately used to initialize the data array.
            Otherwise, if it is an array, it is immediately used as the data
            array, and otherwise the data array is left uninitialized and will
            be initialized by this initializer in :meth:`initialize`. It can
            also be a scalar, in which case the data array will be filled by
            this scalar. Note that float32 is used in this case.
        shape (int or tuple of int or None): Shape of the parameter. If it is
            ``None``, the initialization is deferred to the call of
            :meth:`initialize`.
        name (str): Name of the parameter.

    Attributes:
        initializer: Initializer of the data array. It is used for
            initializing the data array of an uninitialized variable.
        update_rule: :class:`~chainer.optimizer.UpdateRule` instance that
            updates this variable as a parameter. This argument is set to
            :attr:`update_rule`.

    """

    initializer = None  # type: tp.Optional[tp.Union[tp.Optional[types.AbstractInitializer], types.NdArray]] # NOQA
    # TODO(okapies): fix the behavior when shape is None and remove NdArray
    _grad_initializer = None  # type: tp.Optional[types.AbstractInitializer]

    def __init__(self, initializer=None, shape=None, name=None):
        # type: (tp.Optional[types.InitializerSpec], tp.Optional[types.ShapeSpec], tp.Optional[str]) -> None # NOQA

        if initializer is None:
            initializer = constant.NaN()
        elif numpy.isscalar(initializer):
            initializer = constant.Constant(initializer)
        if shape is None:
            if isinstance(initializer, chainer.get_array_types()):
                # parameter initialized by the initial array
                super(Parameter, self).__init__(initializer, name=name)
            else:
                # uninitialized parameter
                super(Parameter, self).__init__(name=name, _grad_valid=False)
                dtype = getattr(initializer, 'dtype', None)
                self._grad_initializer = constant.NaN(dtype)
        else:
            # parameter initialized with a given shape
            if isinstance(initializer, chainer.get_array_types()):
                xp = backend.get_array_module(initializer)
                initializer = constant.Constant(initializer)
            else:
                xp = numpy
            data = initializers.generate_array(initializer, shape, xp)  # type: ignore # NOQA
            grad = xp.full_like(data, numpy.nan)
            super(Parameter, self).__init__(data, name=name, grad=grad)

        self._initial_device = backend.CpuDevice()
        self.update_rule = None
        self.initializer = initializer

    def __copy__(self):
        return self._copy_to(Parameter())

    def __reduce__(self):
        args = (
            self.array, self.name, self._grad, self._grad_valid,
            self.initializer, self.update_rule, self.device)
        return _recover_parameter, args

    @property
    def dtype(self):
        array = self.array
        if array is not None:
            return array.dtype
        # uninitialized
        initializer = self.initializer
        if hasattr(initializer, 'dtype'):
            return numpy.dtype(initializer.dtype)
        raise RuntimeError(
            'Dtype of the parameter is not determined yet because it\'s '
            'uninitialized and dtype was not explicitly given.')

    def to_cpu(self):
        return self.to_device(backend.CpuDevice())

    def to_gpu(self, device=None):
        device = chainer.get_device(cuda._get_device_or_current(device))
        assert device.xp is cuda.cupy
        self.to_device(device)

    def to_intel64(self):
        self.to_device(intel64.Intel64Device())

    def to_chx(self):
        if not chainerx.is_available():
            raise RuntimeError('ChainerX is not available.')

        # Derive the target ChainerX device from the array if it is
        # initialized. Otherwise, from the current initial device.
        if self.array is not None:
            device = backend.get_device_from_array(self.array)
        else:
            device = self._initial_device

        if device.xp is numpy:
            self._initial_device = backend.ChainerxDevice(
                chainerx.get_device('native:0'))
        elif device.xp is cuda.cupy:
            self._initial_device = backend.ChainerxDevice(
                chainerx.get_device('cuda:{}'.format(device.device.id)))

        super(Parameter, self)._to_chx(allow_unchaining=True)

    def from_chx(self):
        if self.array is not None:
            device = backend.get_device_from_array(self.array)
        else:
            device = self._initial_device

        if device.xp is chainerx:
            backend_name = device.device.backend.name
            if backend_name == 'native':
                self._initial_device = backend.CpuDevice()
            elif backend_name == 'cuda':
                self._initial_device = backend.GpuDevice.from_device_id(
                    device.device.index)

        super(Parameter, self)._from_chx(allow_unchaining=True)

    def to_device(self, device):
        device = chainer.get_device(device)
        if self.data is None and self._initial_device != device:
            self._data = [None]  # Renew placeholder to break sharing
            self._has_chainerx_array = False
        self._initial_device = device
        super(Parameter, self)._to_device(device, allow_unchaining=True)

    def cleargrad(self):
        super(Parameter, self).cleargrad()
        if self.array is None:
            self._grad_initializer = None

    def zerograd(self):
        super(Parameter, self).zerograd()
        if self.array is None:
            dtype = getattr(self.initializer, 'dtype', None)
            self._grad_initializer = initializers.Zero(dtype)

    def initialize(self, shape):
        """Initializes the uninitialized variable.

        Uninitialized variable is a variable created with the data array set to
        None. This method creates and initializes the data array. The shape of
        the variable can be left unknown until this method is called.

        Args:
            shape (tuple of int): Shape of the data array.

        """
        device = self._initial_device
        assert device is not None
        xp = device.xp

        data = initializers.generate_array(
            self.initializer, shape, xp, device=device)
        ginit = self._grad_initializer
        grad = None if ginit is None else initializers.generate_array(
            ginit, shape, xp, device=device)

        self.array = data
        self.grad = grad

        # Convert the array for iDeep.
        # TODO(niboshi): This could be done in generate_array().
        if isinstance(self._initial_device, intel64.Intel64Device):
            self.to_intel64()

    def update(self):
        """Updates the data array using the gradient and the update rule.

        This method updates the parameter using the attached update rule.

        """
        if self.update_rule is not None:
            self.update_rule.update(self)


def as_variable(obj):
    """Converts an array or a variable into :class:`~chainer.Variable`.

    This is a convenient function to get a :class:`~chainer.Variable` object
    transparently from a raw array or a variable.

    Note that this function should only be used for type consistency (i.e., to
    enforce the return value of an API having type :class:`~chainer.Variable`).
    The :class:`~chainer.Variable.requires_grad` flag is kept as is; if ``obj``
    is a raw array, the newly created variable has ``requires_grad = False``.
    In order to make a variable w.r.t. which you want to compute the gradient,
    you should use :class:`~chainer.Variable` directly.

    Args:
        obj (:ref:`ndarray` or ~chainer.Variable): An array or
            a variable that you want to convert to :class:`~chainer.Variable`.

    Returns:
        ~chainer.Variable:
        A variable converted from ``obj``. If ``obj`` is a raw array, this is a
        new :class:`~chainer.Variable` object that wraps the array. If ``obj``
        is already a :class:`~chainer.Variable` object, this function returns
        ``obj`` as is.

    """
    if isinstance(obj, Variable):
        return obj

    if isinstance(obj, chainerx.ndarray):
        requires_grad = obj.is_backprop_required()
    else:
        requires_grad = False
    return Variable(obj, requires_grad=requires_grad)


def as_array(obj):
    """Returns the underlying array from a variable or an array.

    This is a convenient function to get the underlying array object
    transparently from an object that could be either a variable or an array.

    Args:
        obj (:ref:`ndarray` or ~chainer.Variable): An array or a variable.

    Returns:
        :ref:`ndarray` or ~chainer.Variable:
        The underlying array object of the argument.

    """
    if isinstance(obj, Variable):
        return obj.array
    return obj


def _recover_parameter(*args):
    if len(args) == 7:
        # latest
        data, name, grad, grad_valid, initializer, update_rule, device = args
    elif len(args) == 6:
        data, name, grad, initializer, update_rule, device = args
        grad_valid = True
    else:
        assert False, len(args)

    p = Parameter(initializer=initializer, name=name)
    p.array = data
    p._grad = grad
    p._grad_valid = grad_valid
    p.update_rule = update_rule
    p.to_device(device)
    return p


class _ChainerxVariableNodeProps(object):

    def __init__(self, x):
        self.shape = x.shape
        self.dtype = x.dtype



from chainer.link_hooks.spectral_normalization import SpectralNormalization  # NOQA
from chainer.link_hooks.timer import TimerHook  # NOQA
from chainer.link_hooks.weight_standardization import WeightStandardization  # NOQA



from __future__ import absolute_import
import collections
import heapq
import warnings

import six

import chainer
from chainer import _backprop_utils
from chainer import backend
from chainer.utils import argument
import chainerx


def backward(outputs, grad_outputs=None, **kwargs):
    """backward(outputs, grad_outputs=None, *, enable_double_backprop=False)

    Runs backpropagation from variables simultaneously.

    .. warning::

        This feature is experimental. The interface can change in the future.

    Args:
        outputs (tuple or list of :class:`~chainer.Variable`):
            A sequence of output variables from which backprop starts.
        grad_outputs (None or tuple or list of :class:`~chainer.Variable`):
            A sequence of variables that gives the initial value of each output
            gradient.
            If this argument is ``None``, backprop uses
            :attr:`~chainer.Variable.grad_var` of ``outputs``.
        enable_double_backprop (bool): If ``True``,
            computational trace of the whole backpropagation procedure is
            recorded to the computational graph so that one can further do
            backpropagation from the resulting gradients. Note that
            enabling it results in larger memory consumption needed to
            store the gradients w.r.t intermediate variables that are
            required for the second gradient computation.

    .. seealso::
       :meth:`chainer.Variable.backward`
       :func:`chainer.grad`

    """
    enable_double_backprop, = argument.parse_kwargs(
        kwargs, ('enable_double_backprop', False),
        retain_grad='semantics for retain_grad=True is under discussion',
        loss_scale='chainer.backward does not support loss_scale option',
    )
    if not isinstance(outputs, (tuple, list)):
        raise TypeError(
            'outputs must be a tuple or a list, not {}.'.format(type(outputs)))
    for v in outputs:
        if not isinstance(v, chainer.Variable):
            raise TypeError(
                'each output must be a Variable, not {}'.format(type(v)))
    if grad_outputs is not None:
        if not isinstance(grad_outputs, (tuple, list)):
            raise TypeError(
                'grad_outputs must be None, a tuple, or a list, not {}.'
                .format(type(grad_outputs)))
        if len(outputs) != len(grad_outputs):
            raise ValueError(
                'grad_outputs must be of the same length as outputs.\n'
                'len(outputs) = {}, len(grad_outputs) = {}'
                .format(len(outputs), len(grad_outputs)))

    is_chainerx = [v._has_chainerx_array for v in outputs]

    if any(is_chainerx):
        if not all(is_chainerx):
            # The restriction is required as soon as the workarounds below
            # are removed.
            raise ValueError('cannot mix chainerx and other backends')

        # Cannot use chainerx.backward directly, because it does not follow
        # retain_grad=False
        # TODO(kataoka): Fix chainerx.backward and remove this workaround
        if grad_outputs is None:
            grad_outputs = []
            for y in outputs:
                grad_outputs.append(y.grad_var)
                y.grad_var = None

        # The check is required because chainerx.backward sets default grads.
        # TODO(kataoka): Fix chainerx.backward and remove this workaround
        indices = [i for i, gy in enumerate(grad_outputs) if gy is not None]
        outputs = [outputs[i] for i in indices]
        grad_outputs = [grad_outputs[i] for i in indices]

        # Use new variables to start backprop
        # TODO(kataoka): Implement chainerx.backward(output, grad_outputs)
        # and remove this workaround.
        outputs = chainer.functions.identity(*outputs)
        if not isinstance(outputs, tuple):
            outputs = outputs,
        grad_outputs = chainer.functions.identity(*grad_outputs)
        if not isinstance(grad_outputs, tuple):
            grad_outputs = grad_outputs,

        # TODO(kataoka): Even after F.identity, non-float grad cannot be set.
        # Move the check to elsewhere and remove this workaround.
        outputs_ = []
        for y, gy in zip(outputs, grad_outputs):
            if not y.requires_grad and gy is not None:
                warnings.warn(
                    'Some of grads are ignored by chainer.backward.\n'
                    'backend: ChainerX, '
                    'output.dtype: {}, grad_output.dtype: {}'.format(
                        y.dtype, gy.dtype),
                    RuntimeWarning)
                continue
            y.grad_var = gy
            outputs_.append(y)
        outputs = outputs_
        del outputs_

        # See also the ChainerX case of Variable.backward
        arrs = []
        for y in outputs:
            arr = y._data[0]
            assert isinstance(arr, chainerx.ndarray)
            arrs.append(arr)
        chainerx.backward(
            arrs, enable_double_backprop=enable_double_backprop)
        return

    if grad_outputs is None:
        grad_outputs = []
        for y in outputs:
            grad_var = y.grad_var
            if grad_var is None:
                warnings.warn(
                    'outputs contains a Variable without grad, or '
                    'duplicate outputs. Note that '
                    'chainer.backward does not set default grad.',
                    RuntimeWarning)
            y.grad_var = None
            grad_outputs.append(grad_var)
    outputs = [
        (y.node, gy) for y, gy in zip(outputs, grad_outputs) if gy is not None]
    with chainer.using_config('enable_backprop', enable_double_backprop):
        _backprop_to_all(outputs, False, None)


def _backprop_to_all(outputs, retain_grad, loss_scale):
    """Backprop to all input variables

    Args:
        outputs (list of tuple): each tuple is (y_node, y_grad_var).
            y_grad_var should not be None.
        retain_grad (bool): see docstring of Variable.backward
        loss_scale (float): see docstring of Variable.backward

    """
    OrderedDict = chainer.utils._collections.OrderedDict  # fix py2 memory leak

    cand_funcs = []
    seen_set = set()

    def add_cand(cand):
        if cand not in seen_set:
            # Negate since heapq is min-heap
            heapq.heappush(cand_funcs, (-cand.rank, len(seen_set), cand))
            seen_set.add(cand)

    grads = _backprop_utils.GradTable(accumulate_grad_inputs=True)

    leaf_nodes = set()

    for y, gy in outputs:
        grads.accumulate(y, gy)

        func = y.creator_node
        if func is None:  # leaf
            leaf_nodes.add(y)
        else:
            add_cand(func)

    # Fix F812 (Python 2)
    y = None
    del y

    is_debug = chainer.is_debug()
    base_hooks = chainer.get_function_hooks().values()
    while cand_funcs:
        _, _, func = heapq.heappop(cand_funcs)
        inputs = func.inputs
        target_input_indexes = tuple([
            i for i, x in enumerate(inputs) if x.requires_grad
        ])
        outputs = [y() for y in func.outputs]  # access via weak ref
        out_grad = tuple([grads.pop(y)
                          if y is not None and y.creator_node is not None
                          else None
                          for y in outputs])
        if not target_input_indexes:
            continue

        in_data = [x.data for x in inputs]
        out_grad_array = [None if g is None else g.array for g in out_grad]
        if func._n_local_function_hooks != 0:
            local_hooks = collections.OrderedDict(chainer.get_function_hooks())
            local_hooks.update(func.local_function_hooks)
            hooks = local_hooks.values()  # avoid six for performance
        else:
            hooks = base_hooks

        with chainer.using_device(
                backend.get_device_from_array(*(in_data + out_grad_array))):
            for hook in hooks:
                hook.backward_preprocess(
                    func, tuple(in_data), tuple(out_grad_array))

            # Collect the current input gradients.
            target_inputs = [inputs[i] for i in target_input_indexes]
            # Keep the order for the portability, rather than
            # in_grad = {x: grads.get_as_list(x)
            #            for x in set(target_inputs)}
            in_grad = OrderedDict()
            for x in target_inputs:
                if x not in in_grad:
                    in_grad[x] = grads.get_as_list(x)

            _backprop_utils.backprop_step(
                func, target_input_indexes, out_grad, in_grad, is_debug)

            for hook in hooks:
                hook.backward_postprocess(
                    func, tuple(in_data), tuple(out_grad_array))

        if retain_grad:
            # The gradients of the outputs of `func` are final. Store them if
            # retain_grad=True.
            for y, gy in six.moves.zip(outputs, out_grad):
                if y is not None:
                    y._set_grad_var_if_available(gy)
            del gy  # to reduce memory usage
        del out_grad  # to reduce memory usage

        for x, gx in in_grad.items():
            if not gx:  # gradient == None
                continue

            for gx_elem in gx:
                if gx_elem is not None:
                    chainer.variable._check_grad_type(
                        func, x, True, gx_elem.array)
            del gx_elem  # to reduce memory usage

            if x.creator_node is None:  # leaf
                leaf_nodes.add(x)
            else:
                add_cand(x.creator_node)
        del gx, in_grad  # to reduce memory usage

    for x in leaf_nodes:
        x_var = x.get_variable_or_none()
        gx = grads.pop(x)
        if x_var is not None:
            x_var._set_grad_var_without_check(gx)
            x_var._loss_scale = loss_scale
    grads.assert_no_grads()



import chainerx


# TODO(sonots): Implement in C++
def clip(a, a_min, a_max):
    """Clips the values of an array to a given interval.

    Given an interval, values outside the interval are clipped to the
    interval edges. For example, if an interval of ``[0, 1]`` is specified,
    values smaller than 0 become 0, and values larger than 1 become 1.

    Args:
        a (~chainerx.ndarray): Array containing elements to clip.
        a_min (scalar): Maximum value.
        a_max (scalar): Minimum value.

    Returns:
        ~chainerx.ndarray: An array with the elements of ``a``, but where
        values < ``a_min`` are replaced with ``a_min``,
        and those > ``a_max`` with ``a_max``.

    Note:
        The :class:`~chainerx.ndarray` typed ``a_min`` and ``a_max`` are
        not supported yet.

    Note:
        During backpropagation, this function propagates the gradient
        of the output array to the input array ``a``.

    .. seealso:: :func:`numpy.clip`

    """
    if a_min is None and a_max is None:
        raise ValueError('Must set either a_min or a_max.')

    if a_min is not None:
        a = chainerx.maximum(a, a_min)

    if a_max is not None:
        a = chainerx.minimum(a, a_max)

    return a



from __future__ import division
import datetime
import sys

from chainer.training import extension
from chainer.training.extensions import util


class ProgressBar(extension.Extension):

    """Trainer extension to print a progress bar and recent training status.

    This extension prints a progress bar at every call. It watches the current
    iteration and epoch to print the bar.

    Args:
        training_length (tuple): Length of whole training. It consists of an
            integer and either ``'epoch'`` or ``'iteration'``. If this value is
            omitted and the stop trigger of the trainer is
            :class:`IntervalTrigger`, this extension uses its attributes to
            determine the length of the training.
        update_interval (int): Number of iterations to skip printing the
            progress bar.
        bar_length (int): Length of the progress bar in characters.
        out: Stream to print the bar. Standard output is used by default.

    """

    def __init__(self, training_length=None, update_interval=100,
                 bar_length=50, out=sys.stdout):
        self._training_length = training_length
        self._status_template = None
        self._update_interval = update_interval
        self._bar_length = bar_length
        self._out = out
        self._recent_timing = []
        self._pbar = _TrainerProgressBar()

    def __call__(self, trainer):
        self._pbar.trainer = trainer

        iteration = trainer.updater.iteration

        # print the progress bar
        if iteration % self._update_interval == 0:
            self._pbar.update()

    def finalize(self):
        self._pbar.close()


class _TrainerProgressBar(util.ProgressBar):

    trainer = None
    training_length = None
    status_template = None

    def get_lines(self):
        lines = []

        iteration = self.trainer.updater.iteration
        epoch = self.trainer.updater.epoch_detail

        if self.training_length is None:
            t = self.trainer.stop_trigger
            self.training_length = t.get_training_length()
        length, unit = self.training_length

        if unit == 'iteration':
            rate = iteration / length
        else:
            rate = epoch / length
        rate = min(rate, 1.0)

        bar_length = self._bar_length
        marks = '#' * int(rate * bar_length)
        lines.append('     total [{}{}] {:6.2%}\n'.format(
            marks, '.' * (bar_length - len(marks)), rate))

        epoch_rate = epoch - int(epoch)
        marks = '#' * int(epoch_rate * bar_length)
        lines.append('this epoch [{}{}] {:6.2%}\n'.format(
            marks, '.' * (bar_length - len(marks)), epoch_rate))

        if self.status_template is None:
            self.status_template = (
                '{0.iteration:10} iter, {0.epoch} epoch / %s %ss\n' %
                self.training_length)
        status = self.status_template.format(self.trainer.updater)
        lines.append(status)

        speed_t, speed_e = self.update_speed(iteration, epoch)

        if unit == 'iteration':
            estimated_time = (length - iteration) / speed_t
        else:
            estimated_time = (length - epoch) / speed_e
        estimated_time = max(estimated_time, 0.0)
        lines.append('{:10.5g} iters/sec. Estimated time to finish: {}.\n'
                     .format(speed_t,
                             datetime.timedelta(seconds=estimated_time)))

        return lines



from chainermn.datasets.empty_dataset import create_empty_dataset  # NOQA
from chainermn.datasets.scatter import DataSizeError  # NOQA
from chainermn.datasets.scatter import scatter_index  # NOQA
from chainermn.datasets.scatter import scatter_dataset  # NOQA



from chainer.dataset.tabular import tabular_dataset


class _AsTuple(tabular_dataset.TabularDataset):

    def __init__(self, dataset):
        self._dataset = dataset

    def __len__(self):
        return len(self._dataset)

    @property
    def keys(self):
        return self._dataset.keys

    @property
    def mode(self):
        return tuple

    def get_examples(self, indices, key_indices):
        return self._dataset.get_examples(indices, key_indices)

    def convert(self, data):
        return self._dataset.convert(data)


class _AsDict(tabular_dataset.TabularDataset):

    def __init__(self, dataset):
        self._dataset = dataset

    def __len__(self):
        return len(self._dataset)

    @property
    def keys(self):
        return self._dataset.keys

    @property
    def mode(self):
        return dict

    def get_examples(self, indices, key_indices):
        return self._dataset.get_examples(indices, key_indices)

    def convert(self, data):
        return self._dataset.convert(data)



from chainer.functions.activation import prelu
from chainer import link
from chainer import variable


class PReLU(link.Link):

    """Parametric ReLU function as a link.

    Args:
        shape (tuple of ints): Shape of the parameter array.
        init (float): Initial parameter value.

    See the paper for details: `Delving Deep into Rectifiers: Surpassing
    Human-Level Performance on ImageNet Classification
    <https://arxiv.org/abs/1502.01852>`_.

    To try PReLU instead of ReLU, replace ``F.relu`` with individual ``PReLU``
    links registered to the model. For example, the model defined in the
    `MNIST example
    <https://github.com/chainer/chainer/tree/master/examples/mnist/train_mnist.py>`_
    can be rewritten as follows.

    ReLU version (original)::

        class MLP(chainer.Chain):

            def __init__(self, n_units, n_out):
                super(MLP, self).__init__()
                with self.init_scope():
                    self.l1 = L.Linear(None, n_units)
                    self.l2 = L.Linear(None, n_units)
                    self.l3 = L.Linear(None, n_out)

            def forward(self, x):
                h1 = F.relu(self.l1(x))
                h2 = F.relu(self.l2(h1))
                return self.l3(h2)

    PReLU version::

        class MLP(chainer.Chain):

            def __init__(self, n_units, n_out):
                super(MLP, self).__init__()
                with self.init_scope():
                    self.l1 = L.Linear(None, n_units)
                    self.a1 = L.PReLU()
                    self.l2 = L.Linear(None, n_units)
                    self.a2 = L.PReLU()
                    self.l3 = L.Linear(None, n_out)

            def forward(self, x):
                h1 = self.a1(self.l1(x))
                h2 = self.a2(self.l2(h1))
                return self.l3(h2)

    .. seealso:: :func:`chainer.functions.prelu`

    Attributes:
        W (~chainer.Parameter): Coefficient of parametric ReLU.

    """

    def __init__(self, shape=(), init=0.25):
        super(PReLU, self).__init__()
        with self.init_scope():
            self.W = variable.Parameter(init, shape)

    def forward(self, x):
        """Applies the parametric ReLU activation function.

        Args:
            x (~chainer.Variable): Input variable.

        Returns:
            ~chainer.Variable: Output of the parametric ReLU function.

        """
        return prelu.prelu(x, self.W)



import chainer
import numpy


class _SynchronizedIterator(chainer.dataset.iterator.Iterator):

    def __init__(self, actual_iterator, communicator):
        if not hasattr(actual_iterator, 'order_sampler'):
            raise ValueError('actual_iterator must have order_sampler')
        else:
            super(_SynchronizedIterator, self).__setattr__(
                'actual_iterator', actual_iterator)

        # Synchronize random seed.
        self.communicator = communicator
        if self.communicator.rank == 0:
            seed = numpy.random.randint(0, 2 ** 32 - 1)
        else:
            seed = None
        seed = self.communicator.bcast_obj(seed, root=0)

        # Random number generator for iterator.
        rng = numpy.random.RandomState(seed)
        self.actual_iterator.order_sampler = \
            chainer.iterators.ShuffleOrderSampler(rng)
        self.actual_iterator.reset()

    def __getattr__(self, attr_name):
        return getattr(self.actual_iterator, attr_name)

    def __setattr__(self, attr_name, value):
        setattr(self.actual_iterator, attr_name, value)

    def __next__(self):
        return self.actual_iterator.__next__()

    def serialize(self, serializer):
        self.actual_iterator.serialize(serializer)


def create_synchronized_iterator(actual_iterator, communicator):
    """Create a synchronized iterator from a Chainer iterator.

    This iterator shares the same batches on multiple processes,
    using the same random number generators to maintain the order of batch
    shuffling same.

    Here is an example situation.
    When we train a sequence-to-sequence model, where the encoder and
    the decoder is located on two different processes, we want to share
    the same batches on each process, thus inputs for the encoder and
    output teacher signals for the decoder become consistent.

    In order to use the synchronized iterator, first create the iterator
    from Chainer iterator and ChainerMN communicator::

        iterator = chainermn.iterators.create_synchronized_iterator(
            chainer.iterators.SerialIterator(
                dataset, batch_size, shuffle=True),
            communicator)

    Then you can use it as the ordinary Chainer iterator::

        updater = chainer.training.StandardUpdater(iterator, optimizer)
        trainer = training.Trainer(updater)
        trainer.run()

    The resulting iterator shares the same shuffling order among processes
    in the specified communicator.

    Args:
        actual_iterator: Chainer iterator
            (e.g., ``chainer.iterators.SerialIterator``).
        communicator: ChainerMN communicator.

    Returns:
        The synchronized iterator based on ``actual_iterator``.
    """
    chainer.utils.experimental(
        'chainermn.iterators.create_synchronized_iterator')

    return _SynchronizedIterator(actual_iterator, communicator)



from chainer import backend
from chainer import function
from chainer.utils import type_check


class R2_score(function.Function):

    def __init__(self, sample_weight, multioutput):
        if sample_weight is not None:
            raise NotImplementedError()
        if multioutput in ['uniform_average', 'raw_values']:
            self.multioutput = multioutput
        else:
            raise ValueError('invalid multioutput argument')

    def check_type_forward(self, in_types):
        type_check._argname(in_types, ('pred', 'true'))
        pred_type, true_type = in_types

        type_check.expect(
            pred_type.dtype.kind == 'f',
            true_type.dtype.kind == 'f'
        )

        type_check.expect(
            pred_type.shape == true_type.shape,
        )

    def forward(self, inputs):
        xp = backend.get_array_module(*inputs)
        pred, true = inputs
        SS_res = xp.asarray(
            xp.sum((pred - true) ** 2, axis=0))
        SS_tot = xp.asarray(
            xp.sum((true - xp.mean(true, axis=0)) ** 2, axis=0))
        SS_tot_iszero = SS_tot == 0
        SS_tot[SS_tot_iszero] = 1  # Assign dummy value to avoid zero-division
        ret = xp.where(
            SS_tot_iszero, 0.0, 1 - SS_res / SS_tot
        ).astype(pred.dtype, copy=False)
        if self.multioutput == 'uniform_average':
            return xp.asarray(ret.mean()),
        elif self.multioutput == 'raw_values':
            return ret,


def r2_score(pred, true, sample_weight=None, multioutput='uniform_average'):
    """Computes R^2(coefficient of determination) regression score function.

    Args:
        pred (:class:`~chainer.Variable` or :ref:`ndarray`): Variable holding a
            vector, matrix or tensor of estimated target values.
        true (:class:`~chainer.Variable` or :ref:`ndarray`): Variable holding a
            vector, matrix or tensor of correct target values.
        sample_weight: This argument is for compatibility with scikit-learn's
                implementation of r2_score. Current implementation admits None
                only.
        multioutput(string): ['uniform_average', 'raw_values']. if
                'uniform_average', this function returns an average of R^2
                score of multiple output. If 'raw_average', this function
                return a set of R^2 score of multiple output.
    Returns:
        ~chainer.Variable: A Variable holding a scalar array of the R^2 score
        if 'multioutput' is 'uniform_average' or a vector of R^2 scores if
        'multioutput' is 'raw_values'.

    .. note:: This function is non-differentiable.

    """
    return R2_score(sample_weight=sample_weight,
                    multioutput=multioutput)(pred, true)



import numpy

from chainer.backends import cuda
from chainer import function_node
from chainer.functions.math import erfcx
from chainer import utils
from chainer.utils import type_check


_log_ndtr_cpu = None


class LogNdtr(function_node.FunctionNode):

    @property
    def label(self):
        return 'log_ndtr'

    def check_type_forward(self, in_types):
        type_check.expect(in_types.size() == 1)
        type_check.expect(in_types[0].dtype.kind == 'f')

    def forward_cpu(self, x):
        global _log_ndtr_cpu
        if _log_ndtr_cpu is None:
            try:
                from scipy import special
                _log_ndtr_cpu = special.log_ndtr
            except ImportError:
                raise ImportError('SciPy is not available. Forward computation'
                                  ' of log_ndtr can not be done.')

        self.retain_inputs((0,))
        return utils.force_array(_log_ndtr_cpu(x[0]), dtype=x[0].dtype),

    def forward_gpu(self, x):
        self.retain_inputs((0,))
        return cuda.elementwise(
            'T x', 'T y',
            '''
            if (x > 0) {
                y = log1p(-normcdf(-x));
            } else {
                y = log(0.5 * erfcx(-sqrt(0.5) * x)) - 0.5 * x * x;
            }
            ''',
            'elementwise_log_ndtr',
        )(x[0]),

    def backward(self, indexes, gy):
        x = self.get_retained_inputs()[0]
        return (2 / numpy.pi) ** 0.5 / erfcx.erfcx(- x / 2 ** 0.5) * gy[0],


def log_ndtr(x):
    """Logarithm of cumulative distribution function of normal distribution.

    .. note::
       Forward computation in CPU can not be done if
       `SciPy <https://www.scipy.org/>`_ is not available.

    Args:
        x (:class:`~chainer.Variable` or :ref:`ndarray`): Input variable.

    Returns:
        ~chainer.Variable: Output variable.
    """
    return LogNdtr().apply((x,))[0]



import contextlib
import typing as tp  # NOQA
import unittest

import numpy
import six

import chainer
from chainer import backend
from chainer import initializers
from chainer.testing import array as array_module
from chainer import utils


class _TestError(AssertionError):

    """Parent class to Chainer test errors."""

    @classmethod
    def check(cls, expr, message):
        if not expr:
            raise cls(message)

    @classmethod
    def fail(cls, message, exc=None):
        if exc is not None:
            utils._raise_from(cls, message, exc)
        raise cls(message)

    @classmethod
    @contextlib.contextmanager
    def raise_if_fail(cls, message, error_types=AssertionError):
        try:
            yield
        except error_types as e:
            cls.fail(message, e)


class FunctionTestError(_TestError):
    """Raised when the target function is implemented incorrectly."""
    pass


class LinkTestError(_TestError):
    """Raised when the target link is implemented incorrectly."""
    pass


class InitializerArgument(object):

    """Class to hold a pair of initializer argument value and actual
    initializer-like.

    This class is meant to be included in the return value from
    :meth:`chainer.testing.LinkTestCase.get_initializers` in
    :class:`chainer.testing.LinkTestCase` if the argument and the actual
    initializer in the link do not directly correspond.
    In that case, the first element should correspond to the argument passed to
    the constructor of the link, and the second element correspond to the
    actual initializer-like object used by the link.
    """

    def __init__(self, argument_value, expected_initializer):
        if expected_initializer is None:
            raise ValueError('Expected initialized cannot be None.')
        initializers._check_is_initializer_like(expected_initializer)

        self.argument_value = argument_value
        self.expected_initializer = expected_initializer


class FunctionTestBase(object):

    backend_config = None
    check_forward_options = None
    check_backward_options = None
    check_double_backward_options = None
    skip_forward_test = False
    skip_backward_test = False
    skip_double_backward_test = False
    dodge_nondifferentiable = False
    numerical_grad_dtype = numpy.float64
    contiguous = None

    def __init__(self, *args, **kwargs):
        super(FunctionTestBase, self).__init__(*args, **kwargs)
        self.check_forward_options = {}
        self.check_backward_options = {}
        self.check_double_backward_options = {}

    def before_test(self, test_name):
        pass

    def forward(self, inputs, device):
        raise NotImplementedError('forward() is not implemented.')

    def forward_expected(self, inputs):
        raise NotImplementedError('forward_expected() is not implemented.')

    def generate_inputs(self):
        raise NotImplementedError('generate_inputs() is not implemented.')

    def generate_grad_outputs(self, outputs_template):
        grad_outputs = tuple([
            numpy.random.uniform(-1, 1, a.shape).astype(a.dtype)
            for a in outputs_template])
        return grad_outputs

    def generate_grad_grad_inputs(self, inputs_template):
        grad_grad_inputs = tuple([
            numpy.random.uniform(-1, 1, a.shape).astype(a.dtype)
            for a in inputs_template])
        return grad_grad_inputs

    def check_forward_outputs(self, outputs, expected_outputs):
        assert isinstance(outputs, tuple)
        assert isinstance(expected_outputs, tuple)
        assert all(isinstance(a, chainer.get_array_types()) for a in outputs)
        assert all(
            isinstance(a, chainer.get_array_types()) for a in expected_outputs)
        _check_arrays_equal(
            outputs, expected_outputs, FunctionTestError,
            **self.check_forward_options)

    def _to_noncontiguous_as_needed(self, contig_arrays):
        if self.contiguous is None:
            # non-contiguous
            return array_module._as_noncontiguous_array(contig_arrays)
        if self.contiguous == 'C':
            # C-contiguous
            return contig_arrays
        assert False, (
            'Invalid value of `contiguous`: {}'.format(self.contiguous))

    def _generate_inputs(self):
        inputs = self.generate_inputs()
        _check_array_types(inputs, backend.CpuDevice(), 'generate_inputs')
        return inputs

    def _generate_grad_outputs(self, outputs_template):
        grad_outputs = self.generate_grad_outputs(outputs_template)
        _check_array_types(
            grad_outputs, backend.CpuDevice(), 'generate_grad_outputs')
        return grad_outputs

    def _generate_grad_grad_inputs(self, inputs_template):
        grad_grad_inputs = self.generate_grad_grad_inputs(inputs_template)
        _check_array_types(
            grad_grad_inputs, backend.CpuDevice(), 'generate_grad_grad_inputs')
        return grad_grad_inputs

    def _forward_expected(self, inputs):
        outputs = self.forward_expected(inputs)
        _check_array_types(
            outputs, backend.CpuDevice(), 'forward_expected')
        return outputs

    def _forward(self, inputs, backend_config):
        assert all(isinstance(a, chainer.Variable) for a in inputs)
        with backend_config:
            outputs = self.forward(inputs, backend_config.device)
        _check_variable_types(
            outputs, backend_config.device, 'forward', FunctionTestError)
        return outputs

    def run_test_forward(self, backend_config):
        # Runs the forward test.

        if self.skip_forward_test:
            raise unittest.SkipTest('skip_forward_test is set')

        self.backend_config = backend_config
        self.test_name = 'test_forward'
        self.before_test(self.test_name)

        cpu_inputs = self._generate_inputs()
        cpu_inputs = self._to_noncontiguous_as_needed(cpu_inputs)
        inputs_copied = [a.copy() for a in cpu_inputs]

        # Compute expected outputs
        cpu_expected = self._forward_expected(cpu_inputs)

        # Compute actual outputs
        inputs = backend_config.get_array(cpu_inputs)
        inputs = self._to_noncontiguous_as_needed(inputs)
        outputs = self._forward(
            tuple([
                chainer.Variable(a, requires_grad=a.dtype.kind == 'f')
                for a in inputs]),
            backend_config)

        # Check inputs has not changed
        indices = []
        for i in range(len(inputs)):
            try:
                array_module.assert_allclose(
                    inputs_copied[i], inputs[i], atol=0, rtol=0)
            except AssertionError:
                indices.append(i)

        if indices:
            FunctionTestError.fail(
                'Input arrays have been modified during forward.\n'
                'Indices of modified inputs: {}\n'
                'Input array shapes and dtypes: {}\n'.format(
                    ', '.join(str(i) for i in indices),
                    utils._format_array_props(inputs)))

        self.check_forward_outputs(
            tuple([var.array for var in outputs]),
            cpu_expected)

    def run_test_backward(self, backend_config):
        # Runs the backward test.

        if self.skip_backward_test:
            raise unittest.SkipTest('skip_backward_test is set')

        # avoid cyclic import
        from chainer import gradient_check

        self.backend_config = backend_config
        self.test_name = 'test_backward'
        self.before_test(self.test_name)

        def f(*args):
            return self._forward(args, backend_config)

        def do_check():
            inputs = self._generate_inputs()
            outputs = self._forward_expected(inputs)
            grad_outputs = self._generate_grad_outputs(outputs)

            inputs = backend_config.get_array(inputs)
            grad_outputs = backend_config.get_array(grad_outputs)
            inputs = self._to_noncontiguous_as_needed(inputs)
            grad_outputs = self._to_noncontiguous_as_needed(grad_outputs)

            with FunctionTestError.raise_if_fail(
                    'backward is not implemented correctly'):
                gradient_check.check_backward(
                    f, inputs, grad_outputs, dtype=self.numerical_grad_dtype,
                    detect_nondifferentiable=self.dodge_nondifferentiable,
                    **self.check_backward_options)

        if self.dodge_nondifferentiable:
            while True:
                try:
                    do_check()
                except gradient_check.NondifferentiableError:
                    continue
                else:
                    break
        else:
            do_check()

    def run_test_double_backward(self, backend_config):
        # Runs the double-backward test.

        if self.skip_double_backward_test:
            raise unittest.SkipTest('skip_double_backward_test is set')

        # avoid cyclic import
        from chainer import gradient_check

        self.backend_config = backend_config
        self.test_name = 'test_double_backward'
        self.before_test(self.test_name)

        def f(*args):
            return self._forward(args, backend_config)

        def do_check():
            inputs = self._generate_inputs()
            outputs = self._forward_expected(inputs)
            grad_outputs = self._generate_grad_outputs(outputs)
            grad_grad_inputs = self._generate_grad_grad_inputs(inputs)

            # Drop ggx corresponding to non-differentiable inputs.
            # Generated `grad_grad_inputs`, the upstream gradients for the
            # double backward test, may contain `None` for omitted gradients.
            # These must be propagated to the gradient check.
            grad_grad_inputs = [
                ggx for ggx in grad_grad_inputs
                if (ggx is None or ggx.dtype.kind == 'f')]

            inputs = backend_config.get_array(inputs)
            grad_outputs = backend_config.get_array(grad_outputs)
            grad_grad_inputs = backend_config.get_array(grad_grad_inputs)
            inputs = self._to_noncontiguous_as_needed(inputs)
            grad_outputs = self._to_noncontiguous_as_needed(grad_outputs)
            grad_grad_inputs = (
                self._to_noncontiguous_as_needed(grad_grad_inputs))

            with backend_config:
                with FunctionTestError.raise_if_fail(
                        'double backward is not implemented correctly'):
                    gradient_check.check_double_backward(
                        f, inputs, grad_outputs, grad_grad_inputs,
                        dtype=self.numerical_grad_dtype,
                        detect_nondifferentiable=self.dodge_nondifferentiable,
                        **self.check_double_backward_options)

        if self.dodge_nondifferentiable:
            while True:
                try:
                    do_check()
                except gradient_check.NondifferentiableError:
                    continue
                else:
                    break
        else:
            do_check()


class FunctionTestCase(FunctionTestBase, unittest.TestCase):
    """A base class for function test cases.

    Function test cases can inherit from this class to define a set of function
    tests.

    .. rubric:: Required methods

    Each concrete class must at least override the following three methods.

    ``forward(self, inputs, device)``
        Implements the target forward function.
        ``inputs`` is a tuple of :class:`~chainer.Variable`\\ s.
        This method is expected to return the output
        :class:`~chainer.Variable`\\ s with the same array types as the inputs.
        ``device`` is the device corresponding to the input arrays.

    ``forward_expected(self, inputs)``
        Implements the expectation of the target forward function.
        ``inputs`` is a tuple of :class:`numpy.ndarray`\\ s.
        This method is expected to return the output
        :class:`numpy.ndarray`\\ s.

    ``generate_inputs(self)``
        Returns a tuple of input arrays of type :class:`numpy.ndarray`.

    .. rubric:: Optional methods

    Additionally the concrete class can override the following methods.

    ``before_test(self, test_name)``
        A callback method called before each test.
        Typically a skip logic is implemented by conditionally raising
        :class:`unittest.SkipTest`.
        ``test_name`` is one of ``'test_forward'``, ``'test_backward'``, and
        ``'test_double_backward'``.

    ``generate_grad_outputs(self, outputs_template)``
        Returns a tuple of output gradient arrays of type
        :class:`numpy.ndarray` or ``None`` for omitted the gradients.
        ``outputs_template`` is a tuple of template arrays. The returned arrays
        are expected to have the same shapes and dtypes as the template arrays.

    ``generate_grad_grad_inputs(self, inputs_template)``
        Returns a tuple of the second order input gradient arrays of type
        :class:`numpy.ndarray` or ``None`` for omitted gradients.
        ``input_template`` is a tuple of template arrays. The returned arrays
        are expected to have the same shapes and dtypes as the template arrays.

    ``check_forward_outputs(self, outputs, expected_outputs)``
        Implements check logic of forward outputs. Typically additional check
        can be done after calling ``super().check_forward_outputs``.
        ``outputs`` and ``expected_outputs`` are tuples of arrays.
        In case the check fails, ``FunctionTestError`` should be raised.

    .. rubric:: Configurable attributes

    The concrete class can override the following attributes to control the
    behavior of the tests.

    ``skip_forward_test`` (bool):
        Whether to skip forward computation test. ``False`` by default.

    ``skip_backward_test`` (bool):
        Whether to skip backward computation test. ``False`` by default.

    ``skip_double_backward_test`` (bool):
        Whether to skip double-backward computation test. ``False`` by default.

    ``dodge_nondifferentiable`` (bool):
        Enable non-differentiable point detection in numerical gradient
        calculation. If the inputs returned by ``generate_inputs`` turns
        out to be a non-differentiable point, the test will repeatedly resample
        inputs until a differentiable point will be finally sampled.
        ``False`` by default.

    ``numerical_grad_dtype`` (dtype):
        Input arrays are casted to this dtype when calculating the numerical
        gradients. It is ``float64`` by default, no matter what the original
        input dtypes were, to maximize precision.

    ``contiguous`` (None or 'C'):
        Specifies the contiguousness of incoming arrays (i.e. inputs, output
        gradients, and the second order input gradients). If ``None``, the
        arrays will be non-contiguous as long as possible. If ``'C'``, the
        arrays will be C-contiguous. ``None`` by default.

    .. rubric:: Passive attributes

    These attributes are automatically set.

    ``test_name`` (str):
        The name of the test being run. It is one of ``'test_forward'``,
        ``'test_backward'``, and ``'test_double_backward'``.

    ``backend_config`` (:class:`~chainer.testing.BackendConfig`):
        The backend configuration.

    .. note::

       This class assumes :func:`chainer.testing.inject_backend_tests`
       is used together. See the example below.

    .. admonition:: Example

       .. testcode::

          @chainer.testing.inject_backend_tests(
              None,
              [
                  {},  # CPU
                  {'use_cuda': True},  # GPU
              ])
          class TestReLU(chainer.testing.FunctionTestCase):

              # ReLU function has a non-differentiable point around zero, so
              # dodge_nondifferentiable should be set to True.
              dodge_nondifferentiable = True

              def generate_inputs(self):
                  x = numpy.random.uniform(-1, 1, (2, 3)).astype(numpy.float32)
                  return x,

              def forward(self, inputs, device):
                  x, = inputs
                  return F.relu(x),

              def forward_expected(self, inputs):
                  x, = inputs
                  expected = x.copy()
                  expected[expected < 0] = 0
                  return expected,

    .. seealso:: :class:`~chainer.testing.LinkTestCase`

    """

    def test_forward(self, backend_config):
        """Tests forward computation."""
        self.run_test_forward(backend_config)

    def test_backward(self, backend_config):
        """Tests backward computation."""
        self.run_test_backward(backend_config)

    def test_double_backward(self, backend_config):
        """Tests double-backward computation."""
        self.run_test_double_backward(backend_config)


class _LinkTestBase(object):

    backend_config = None
    contiguous = None

    # List of parameter names represented as strings.
    # I.e. ('gamma', 'beta') for BatchNormalization.
    param_names = ()

    def before_test(self, test_name):
        pass

    def generate_params(self):
        raise NotImplementedError('generate_params is not implemented.')

    def generate_inputs(self):
        raise NotImplementedError('generate_inputs is not implemented.')

    def create_link(self, initializers):
        raise NotImplementedError('create_link is not implemented.')

    def forward(self, link, inputs, device):
        outputs = link(*inputs)
        if not isinstance(outputs, tuple):
            outputs = outputs,
        return outputs

    def check_forward_outputs(self, outputs, expected_outputs):
        assert isinstance(outputs, tuple)
        assert isinstance(expected_outputs, tuple)
        assert all(isinstance(a, chainer.get_array_types()) for a in outputs)
        assert all(
            isinstance(a, chainer.get_array_types()) for a in expected_outputs)
        _check_arrays_equal(
            outputs, expected_outputs, LinkTestError,
            **self.check_forward_options)

    def _generate_params(self):
        params_init = self.generate_params()
        if not isinstance(params_init, (tuple, list)):
            raise TypeError(
                '`generate_params` must return a tuple or a list.')
        for init in params_init:
            _check_generated_initializer(init)
        return params_init

    def _generate_inputs(self):
        inputs = self.generate_inputs()

        _check_array_types(inputs, backend.CpuDevice(), 'generate_inputs')

        return inputs

    def _create_link(self, initializers, backend_config):
        link = self.create_link(initializers)
        if not isinstance(link, chainer.Link):
            raise TypeError(
                '`create_link` must return a chainer.Link object.')

        link.to_device(backend_config.device)

        return link

    def _create_initialized_link(self, inits, backend_config):
        inits = [_get_initializer_argument_value(i) for i in inits]
        link = self._create_link(inits, backend_config)

        # Generate inputs and compute a forward pass to initialize the
        # parameters.
        inputs_np = self._generate_inputs()
        inputs_xp = backend_config.get_array(inputs_np)
        inputs_xp = self._to_noncontiguous_as_needed(inputs_xp)
        input_vars = [chainer.Variable(i) for i in inputs_xp]
        output_vars = self._forward(link, input_vars, backend_config)
        outputs_xp = [v.array for v in output_vars]

        link.cleargrads()

        return link, inputs_xp, outputs_xp

    def _forward(self, link, inputs, backend_config):
        assert all(isinstance(x, chainer.Variable) for x in inputs)

        with backend_config:
            outputs = self.forward(link, inputs, backend_config.device)
        _check_variable_types(
            outputs, backend_config.device, 'forward', LinkTestError)

        return outputs

    def _to_noncontiguous_as_needed(self, contig_arrays):
        if self.contiguous is None:
            # non-contiguous
            return array_module._as_noncontiguous_array(contig_arrays)
        if self.contiguous == 'C':
            # C-contiguous
            return contig_arrays
        assert False, (
            'Invalid value of `contiguous`: {}'.format(self.contiguous))


class LinkTestCase(_LinkTestBase, unittest.TestCase):

    """A base class for link forward and backward test cases.

    Link test cases can inherit from this class to define a set of link tests
    for forward and backward computations.

    .. rubric:: Required methods

    Each concrete class must at least override the following methods.

    ``generate_params(self)``
        Returns a tuple of initializers-likes. The tuple should contain an
        initializer-like for each initializer-like argument, i.e. the
        parameters to the link constructor. These will be passed to
        ``create_link``.

    ``create_link(self, initializers)``
        Returns a link. The link should be initialized with the given
        initializer-likes ``initializers``. ``initializers`` is a tuple of
        same length as the number of parameters.

    ``generate_inputs(self)``
        Returns a tuple of input arrays of type :class:`numpy.ndarray`.

    ``forward(self, link, inputs, device)``
        Implements the target forward function.
        ``link`` is a link created by ``create_link`` and
        ``inputs`` is a tuple of :class:`~chainer.Variable`\\ s.
        This method is expected to return the output
        :class:`~chainer.Variable`\\ s with the same array types as the inputs.
        ``device`` is the device corresponding to the input arrays.
        A default implementation is provided for links that only takes the
        inputs defined in ``generate_inputs`` (wrapped in
        :class:`~chainer.Variable`\\ s) and returns nothing but output
        :class:`~chainer.Variable`\\ s in its forward computation.

    .. rubric:: Optional methods

    Each concrete class may override the following methods depending on the
    skip flags  ``skip_forward_test`` and  ``skip_backward_test``.

    ``before_test(self, test_name)``
        A callback method called before each test.
        Typically a skip logic is implemented by conditionally raising
        :class:`unittest.SkipTest`.
        ``test_name`` is one of ``'test_forward'`` and ``'test_backward'``.

    ``forward_expected(self, link, inputs)``
        Implements the expectation of the target forward function.
        ``link`` is the initialized link that was used to compute the actual
        forward which the results of this method will be compared against.
        The link is guaranteed to reside on the CPU.
        ``inputs`` is a tuple of :class:`numpy.ndarray`\\ s.
        This method is expected to return the output
        :class:`numpy.ndarray`\\ s.
        This method must be implemented if either ``skip_forward_test`` or
        ``skip_backward_test`` is ``False`` in which case forward or backward
        tests are executed.

    ``generate_grad_outputs(self, outputs_template)``
        Returns a tuple of output gradient arrays of type
        :class:`numpy.ndarray`.
        ``outputs_template`` is a tuple of template arrays. The returned arrays
        are expected to have the same shapes and dtypes as the template arrays.

    ``check_forward_outputs(self, outputs, expected_outputs)``
        Implements check logic of forward outputs. Typically additional check
        can be done after calling ``super().check_forward_outputs``.
        ``outputs`` and ``expected_outputs`` are tuples of arrays.
        In case the check fails, ``LinkTestError`` should be raised.

    .. rubric:: Attributes

    The concrete class can override the following attributes to control the
    behavior of the tests.

    ``param_names`` (tuple of str):
        A tuple of strings with all the names of the parameters that should be
        tested. E.g. ``('gamma', 'beta')`` for the batch normalization link.
        ``()`` by default.

    ``skip_forward_test`` (bool):
        Whether to skip forward computation test. ``False`` by default.

    ``skip_backward_test`` (bool):
        Whether to skip backward computation test. ``False`` by default.

    ``dodge_nondifferentiable`` (bool):
        Enable non-differentiable point detection in numerical gradient
        calculation. If the data returned by
        ``generate_params``, ``create_link`` and ``generate_inputs`` turns out
        to be a non-differentiable point, the test will repeatedly resample
        those until a differentiable point will be finally sampled. ``False``
        by default.

    ``numerical_grad_dtype`` (dtype):
        Input arrays are casted to this dtype when calculating the numerical
        gradients. It is ``float64`` by default, no matter what the original
        input dtypes were, to maximize precision.

    ``contiguous`` (None or 'C'):
        Specifies the contiguousness of incoming arrays (i.e. inputs,
        parameters and gradients. If ``None``, the
        arrays will be non-contiguous as long as possible. If ``'C'``, the
        arrays will be C-contiguous. ``None`` by default.

    .. note::

        This class assumes :func:`chainer.testing.inject_backend_tests`
        is used together. See the example below.

    .. note::

        When implementing :class:`~chainer.testing.LinkTestCase` and
        :class:`~chainer.testing.LinkInitializersTestCase` to test both
        forward/backward and initializers, it is often convenient to refactor
        out common logic in a separate class.

    .. admonition:: Example

        .. testcode::

            @chainer.testing.inject_backend_tests(
              None,
              [
                  {},  # CPU
                  {'use_cuda': True},  # GPU
              ])
            class TestLinear(chainer.testing.LinkTestCase):

                param_names = ('W', 'b')

                def generate_params(self):
                    initialW = numpy.random.uniform(
                        -1, 1, (3, 2)).astype(numpy.float32)
                    initial_bias = numpy.random.uniform(
                        -1, 1, (3,)).astype(numpy.float32)
                    return initialW, initial_bias

                def generate_inputs(self):
                    x = numpy.random.uniform(
                        -1, 1, (1, 2)).astype(numpy.float32)
                    return x,

                def create_link(self, initializers):
                    initialW, initial_bias = initializers
                    link = chainer.links.Linear(
                        2, 3, initialW=initialW, initial_bias=initial_bias)
                    return link

                def forward(self, link, inputs, device):
                    x, = inputs
                    return link(x),

                def forward_expected(self, link, inputs):
                    W = link.W.array
                    b = link.b.array
                    x, = inputs
                    expected = x.dot(W.T) + b
                    return expected,

    .. seealso::
        :class:`~chainer.testing.LinkInitializersTestCase`
        :class:`~chainer.testing.FunctionTestCase`

    """

    check_forward_options = None
    check_backward_options = None
    skip_forward_test = False
    skip_backward_test = False
    dodge_nondifferentiable = False
    numerical_grad_dtype = numpy.float64

    def __init__(self, *args, **kwargs):
        self.check_forward_options = {}
        self.check_backward_options = {}

        super(LinkTestCase, self).__init__(*args, **kwargs)

    def forward_expected(self, link, inputs):
        raise NotImplementedError('forward_expected() is not implemented.')

    def generate_grad_outputs(self, outputs_template):
        grad_outputs = tuple([
            numpy.random.uniform(-1, 1, a.shape).astype(a.dtype)
            for a in outputs_template])
        return grad_outputs

    def test_forward(self, backend_config):
        """Tests forward computation."""

        if self.skip_forward_test:
            raise unittest.SkipTest('skip_forward_test is set')

        self.backend_config = backend_config
        self.before_test('test_forward')

        inits = self._generate_params()
        link = self._create_link(inits, backend_config)

        inputs_np = self._generate_inputs()
        inputs_xp = backend_config.get_array(inputs_np)
        inputs_xp = self._to_noncontiguous_as_needed(inputs_xp)
        input_vars = tuple([chainer.Variable(i) for i in inputs_xp])
        # Compute forward of the link and initialize its parameters.
        output_vars = self._forward(link, input_vars, backend_config)
        outputs_xp = [v.array for v in output_vars]

        # Expected outputs are computed on the CPU so the link must be
        # transferred.
        link.to_device(backend.CpuDevice())

        expected_outputs_np = self._forward_expected(link, inputs_np)

        self.check_forward_outputs(
            tuple(outputs_xp), expected_outputs_np)

    def test_backward(self, backend_config):
        """Tests backward computation."""

        if self.skip_backward_test:
            raise unittest.SkipTest('skip_backward_test is set')

        self.backend_config = backend_config
        self.before_test('test_backward')

        # avoid cyclic import
        from chainer import gradient_check

        def do_check():
            # Generate an initialized temporary link that is already forward
            # propagated. This link is only used to generate necessary data,
            # i.e. inputs, outputs and parameters for the later gradient check
            # and the link itself will be discarded.
            inits = self._generate_params()
            link, inputs, outputs = self._create_initialized_link(
                inits, backend_config)

            # Extract the parameter ndarrays from the initialized link.
            params = _get_link_params(link, self.param_names)
            params = [p.array for p in params]

            # Prepare inputs, outputs and upstream gradients for the gradient
            # check.
            cpu_device = backend.CpuDevice()
            outputs = [cpu_device.send(output) for output in outputs]
            grad_outputs = self._generate_grad_outputs(outputs)
            grad_outputs = backend_config.get_array(grad_outputs)

            inputs = self._to_noncontiguous_as_needed(inputs)
            params = self._to_noncontiguous_as_needed(params)
            grad_outputs = self._to_noncontiguous_as_needed(grad_outputs)

            # Create the link used for the actual forward propagation in the
            # gradient check.
            forward_link, _, _ = self._create_initialized_link(
                inits, backend_config)

            def forward(inputs, ps):

                # Use generated parameters.
                with forward_link.init_scope():
                    for param_name, p in zip(self.param_names, ps):
                        setattr(forward_link, param_name, p)

                return self._forward(forward_link, inputs, backend_config)

            with LinkTestError.raise_if_fail(
                    'backward is not implemented correctly'):
                gradient_check._check_backward_with_params(
                    forward, inputs, grad_outputs, params=params,
                    dtype=self.numerical_grad_dtype,
                    detect_nondifferentiable=self.dodge_nondifferentiable,
                    **self.check_backward_options)

        if self.dodge_nondifferentiable:
            while True:
                try:
                    do_check()
                except gradient_check.NondifferentiableError:
                    continue
                else:
                    break
        else:
            do_check()

    def _forward_expected(self, link, inputs):
        assert all(isinstance(x, numpy.ndarray) for x in inputs)

        outputs = self.forward_expected(link, inputs)
        _check_array_types(inputs, backend.CpuDevice(), 'test_forward')

        return outputs

    def _generate_grad_outputs(self, outputs_template):
        assert all(isinstance(x, numpy.ndarray) for x in outputs_template)

        grad_outputs = self.generate_grad_outputs(outputs_template)
        _check_array_types(
            grad_outputs, backend.CpuDevice(), 'generate_grad_outputs')

        return grad_outputs


class LinkInitializersTestCase(_LinkTestBase, unittest.TestCase):

    """A base class for link parameter initializer test cases.

    Link test cases can inherit from this class to define a set of link tests
    for parameter initialization.

    .. rubric:: Required methods

    Each concrete class must at least override the following methods.

    ``generate_params(self)``
        Returns a tuple of initializers-likes. The tuple should contain an
        initializer-like for each initializer-like argument, i.e. the
        parameters to the link constructor. These will be passed to
        ``create_link``.

    ``create_link(self, initializers)``
        Returns a link. The link should be initialized with the given
        initializer-likes ``initializers``. ``initializers`` is a tuple of
        same length as the number of parameters.

    ``generate_inputs(self)``
        Returns a tuple of input arrays of type :class:`numpy.ndarray`.

    ``forward(self, link, inputs, device)``
        Implements the target forward function.
        ``link`` is a link created by ``create_link`` and
        ``inputs`` is a tuple of :class:`~chainer.Variable`\\ s.
        This method is expected to return the output
        :class:`~chainer.Variable`\\ s with the same array types as the inputs.
        ``device`` is the device corresponding to the input arrays.
        A default implementation is provided for links that only takes the
        inputs defined in ``generate_inputs`` (wrapped in
        :class:`~chainer.Variable`\\ s) and returns nothing but output
        :class:`~chainer.Variable`\\ s in its forward computation.

    ``get_initializers(self)``
        Returns a tuple with the same length as the number of initializers that
        the constructor of the link accepts. Each element in the tuple is a
        container itself, listing all initializers-likes that should be tested.
        Each initializer-like in the tuple is tested one at a time by being
        passed to ``create_link``. When the length of the tuple is greater than
        one (i.e. if the link accepts multiple initializers), the ones not
        being tested are replaced by the ones returned by `generate_params`.
        Initializer-likes returned here should be deterministic since test will
        invoke them multiple times to test the correctness.

        For testing initializer arguments that can be non-initializer values
        such as ``None``, one can use the ``InitializerArgument``, defining a
        pair of the link constructor argument and actual initializer-like used
        by the link.
        This method must be implemented if ``skip_initializers_test`` is
        ``False`` in which case the initializers test is executed.

    .. rubric:: Optional methods

    Each concrete class may override the following methods.

    ``before_test(self, test_name)``
        A callback method called before each test.
        Typically a skip logic is implemented by conditionally raising
        :class:`unittest.SkipTest`.
        ``test_name`` is always of ``'test_initializers'``.

    .. rubric:: Attributes

    The concrete class can override the following attributes to control the
    behavior of the tests.

    ``param_names`` (list of str):
        A list of strings with all the names of the parameters that should be
        tested. E.g. ``['gamma', 'beta']`` for the batch normalization link.
        ``[]`` by default.

    ``contiguous`` (None or 'C'):
        Specifies the contiguousness of incoming arrays (i.e. inputs,
        parameters and gradients. If ``None``, the
        arrays will be non-contiguous as long as possible. If ``'C'``, the
        arrays will be C-contiguous. ``None`` by default.

    .. note::

        This class assumes :func:`chainer.testing.inject_backend_tests`
        is used together. See the example below.

    .. note::

        When implementing :class:`~chainer.testing.LinkTestCase` and
        :class:`~chainer.testing.LinkInitializersTestCase` to test both
        forward/backward and initializers, it is often convenient to refactor
        out common logic in a separate class.

    .. admonition:: Example

        .. testcode::

            @chainer.testing.inject_backend_tests(
              None,
              [
                  {},  # CPU
                  {'use_cuda': True},  # GPU
              ])
            class TestLinear(chainer.testing.LinkInitializersTestCase):

                param_names = ['W', 'b']

                def generate_params(self):
                    initialW = numpy.random.uniform(
                        -1, 1, (3, 2)).astype(numpy.float32)
                    initial_bias = numpy.random.uniform(
                        -1, 1, (3,)).astype(numpy.float32)
                    return initialW, initial_bias

                def generate_inputs(self):
                    x = numpy.random.uniform(
                        -1, 1, (1, 2)).astype(numpy.float32)
                    return x,

                def create_link(self, initializers):
                    initialW, initial_bias = initializers
                    link = chainer.links.Linear(
                        2, 3, initialW=initialW, initial_bias=initial_bias)
                    return link

                def forward(self, link, inputs, device):
                    x, = inputs
                    return link(x),

                def get_initializers(self):
                    initialW = [initializers.Constant(1), 2]
                    initial_bias = [initializers.Constant(2), 3,
                        chainer.testing.link.InitializerArgument(None, 0)]
                    return initialW, initial_bias

    .. seealso::
        :class:`~chainer.testing.LinkTestCase`
        :class:`~chainer.testing.FunctionTestCase`

    """

    check_initializers_options = None

    def __init__(self, *args, **kwargs):
        self.check_initializers_options = {}

        super(LinkInitializersTestCase, self).__init__(*args, **kwargs)

    def get_initializers(self):
        raise NotImplementedError('get_initializers is not implemented.')

    def test_initializers(self, backend_config):
        """Tests that the parameters of a links are correctly initialized."""

        self.backend_config = backend_config
        self.before_test('test_initializers')

        params_inits = self._get_initializers()

        # TODO(hvy): Reduce the number of loop iterations by checking
        # multiple parameters simultaneously.
        for i_param, param_inits in enumerate(params_inits):
            # When testing an initializer for a particular parameter, other
            # initializers are picked from generate_params.
            inits = self._generate_params()
            inits = list(inits)

            for init in param_inits:
                inits[i_param] = init
                self._test_single_initializer(i_param, inits, backend_config)

    def _get_initializers(self):
        params_inits = self.get_initializers()
        if not isinstance(params_inits, (tuple, list)):
            raise TypeError(
                '`get_initializers` must return a tuple or a list.')
        for param_inits in params_inits:
            if not isinstance(param_inits, (tuple, list)):
                raise TypeError(
                    '`get_initializers` must return a tuple or a list of '
                    'tuples or lists.')
            for init in param_inits:
                _check_generated_initializer(init)
        return params_inits

    def _test_single_initializer(self, i_param, inits, backend_config):
        # Given a set of initializer constructor arguments for the link, create
        # and initialize a link with those arguments. `i_param` holds the index
        # of the argument that should be tested among these.
        inits_orig = inits
        inits = [_get_initializer_argument_value(i) for i in inits]
        link, _, _ = self._create_initialized_link(inits, backend_config)

        # Extract the parameters from the initialized link.
        params = _get_link_params(link, self.param_names)

        # Convert the parameter of interest into a NumPy ndarray.
        cpu_device = backend.CpuDevice()
        param = params[i_param]
        param_xp = param.array
        param_np = cpu_device.send(param_xp)

        # The expected values of the parameter is decided by the given
        # initializer. If the initializer is `None`, it should have been
        # wrapped in a InitializerArgument along with the expected initializer
        # that the link should default to in case of `None`.
        #
        # Note that for this to work, the expected parameter must be inferred
        # deterministically.
        expected_init = _get_expected_initializer(inits_orig[i_param])
        expected_np = numpy.empty_like(param_np)
        expected_init(expected_np)

        # Compare the values of the expected and actual parameter.
        _check_arrays_equal(
            (expected_np,), (param_np,), LinkTestError,
            **self.check_initializers_options)


def _check_generated_initializer(init):
    if isinstance(init, InitializerArgument):
        init = init.expected_initializer
    elif init is None:
        raise ValueError(
            'A None initializer must be wrapped in a InitializerArgument '
            'along with the expected initializer fallen back to.')
    initializers._check_is_initializer_like(init)


def _get_initializer_argument_value(init):
    # Returns the initializer that should be passed to the link constructor.

    if isinstance(init, InitializerArgument):
        return init.argument_value
    return init


def _get_expected_initializer(init):
    # Returns the expected initializer for the given initializer.

    if isinstance(init, InitializerArgument):
        init = init.expected_initializer

    assert init is not None

    if not isinstance(init, chainer.Initializer):
        init = chainer.initializers._get_initializer(init)
    return init


def _get_link_params(link, param_names):
    params = []
    for name in param_names:
        param = getattr(link, name, None)
        if param is None:
            raise LinkTestError.fail(
                'Link does not have a parameter named \'{}\'.'.format(name))
        params.append(param)
    return params


def _check_array_types(arrays, device, func_name):
    if not isinstance(arrays, tuple):
        raise TypeError(
            '`{}()` must return a tuple, '
            'not {}.'.format(func_name, type(arrays)))
    if not all(
            a is None or isinstance(a, device.supported_array_types)
            for a in arrays):
        raise TypeError(
            '{}() must return a tuple of arrays supported by device {} or'
            ' None.\nActual: {}'.format(
                func_name, device, tuple([type(a) for a in arrays])))


def _check_variable_types(vars, device, func_name, test_error_cls):
    assert issubclass(test_error_cls, _TestError)

    if not isinstance(vars, tuple):
        test_error_cls.fail(
            '`{}()` must return a tuple, '
            'not {}.'.format(func_name, type(vars)))
    if not all(isinstance(a, chainer.Variable) for a in vars):
        test_error_cls.fail(
            '{}() must return a tuple of Variables.\n'
            'Actual: {}'.format(
                func_name, ', '.join(str(type(a)) for a in vars)))
    if not all(isinstance(a.array, device.supported_array_types)
               for a in vars):
        test_error_cls.fail(
            '{}() must return a tuple of Variables of arrays supported by '
            'device {}.\n'
            'Actual: {}'.format(
                func_name, device,
                ', '.join(str(type(a.array)) for a in vars)))


def _check_arrays_equal(
        actual_arrays, expected_arrays, test_error_cls, **opts):
    # `opts` is passed through to `testing.assert_all_close`.
    # Check all outputs are equal to expected values
    assert issubclass(test_error_cls, _TestError)

    message = None
    detail_message = None
    while True:
        # Check number of arrays
        if len(actual_arrays) != len(expected_arrays):
            message = (
                'Number of outputs ({}, {}) does not match'.format(
                    len(actual_arrays), len(expected_arrays)))
            break

        # Check dtypes and shapes
        dtypes_match = all([
            y.dtype == ye.dtype
            for y, ye in zip(actual_arrays, expected_arrays)])
        shapes_match = all([
            y.shape == ye.shape
            for y, ye in zip(actual_arrays, expected_arrays)])
        if not (shapes_match and dtypes_match):
            message = 'Shapes and/or dtypes do not match'
            break

        # Check values
        errors = []
        for i, (actual, expected) in (
                enumerate(zip(actual_arrays, expected_arrays))):
            try:
                array_module.assert_allclose(actual, expected, **opts)
            except AssertionError as e:
                errors.append((i, e))
        if errors:
            message = (
                'Outputs do not match the expected values.\n'
                'Indices of outputs that do not match: {}'.format(
                    ', '.join(str(i) for i, e in errors)))
            f = six.StringIO()
            for i, e in errors:
                f.write('Error details of output [{}]:\n'.format(i))
                f.write(str(e))
                f.write('\n')
            detail_message = f.getvalue()
            break
        break

    if message is not None:
        msg = (
            '{}\n'
            'Expected shapes and dtypes: {}\n'
            'Actual shapes and dtypes:   {}\n'.format(
                message,
                utils._format_array_props(expected_arrays),
                utils._format_array_props(actual_arrays)))
        if detail_message is not None:
            msg += '\n\n' + detail_message
        test_error_cls.fail(msg)



import functools
import traceback
import warnings

import numpy
import pytest

import chainerx
from chainerx.testing import array


# A test returning this object will have its return value ignored.
#
# This is e.g. useful when a combination of parametrizations and operations
# unintentionally cover non-supported function calls.
# For instance, you might parametrize over shapes (tuples) which are unpacked
# and passed to a function.
# While you might want to test empty tuples for module functions, they should
# maybe be ignored for ndarray functions.
#
# If either chainerx or numpy returns this object, the other module should too.
# Otherwise, the test will be considered inconsistent and be treated as a
# failure.
_ignored_result = object()


# A wrapper to obtain the ignore object.
def ignore():
    return _ignored_result


class _ResultsCheckFailure(Exception):
    def __init__(self, msg, indices, condense_results_func=None):
        self.msg = msg
        self.indices = tuple(indices)
        if condense_results_func is None:
            def condense_results_func(np_r, chx_r):
                return 'chainerx: {} numpy: {}'.format(chx_r, np_r)
        self.condense_results_func = condense_results_func

    def condense_results(self, numpy_result, chainerx_result):
        # Generates a condensed error message for a pair of lowest-level numpy
        # and chainerx results.
        return self.condense_results_func(numpy_result, chainerx_result)


def _call_func(impl, args, kw):
    try:
        result = impl(*args, **kw)
        error = None
        tb = None
    except Exception as e:
        result = None
        error = e
        tb = traceback.format_exc()

    return result, error, tb


def _check_chainerx_numpy_error(chainerx_error, chainerx_tb, numpy_error,
                                numpy_tb, accept_error=()):
    # TODO(sonots): Change error class names of ChainerX to be similar with
    # NumPy, and check names.
    if chainerx_error is None and numpy_error is None:
        pytest.fail(
            'Both chainerx and numpy are expected to raise errors, but not')
    elif chainerx_error is None:
        pytest.fail('Only numpy raises error\n\n' + numpy_tb)
    elif numpy_error is None:
        pytest.fail('Only chainerx raises error\n\n' + chainerx_tb)
    elif not (isinstance(chainerx_error, accept_error) and
              isinstance(numpy_error, accept_error)):
        msg = '''Both chainerx and numpy raise exceptions

chainerx
%s
numpy
%s
''' % (chainerx_tb, numpy_tb)
        pytest.fail(msg)


def _is_numpy_type(result):
    return isinstance(result, (numpy.ndarray, numpy.generic))


def _check_chainerx_numpy_result_array(
        check_result_func, chainerx_result, numpy_result, indices):
    # Compares `chainerx_result` and `numpy_result` as arrays.

    is_chainerx_valid_type = isinstance(chainerx_result, chainerx.ndarray)
    is_numpy_valid_type = _is_numpy_type(numpy_result)

    if not (is_chainerx_valid_type and is_numpy_valid_type):
        raise _ResultsCheckFailure(
            'Using decorator without returning ndarrays. '
            'If you want to explicitly ignore certain tests, '
            'return chainerx.testing.ignore() to avoid this error', indices)

    if chainerx_result.shape != numpy_result.shape:
        raise _ResultsCheckFailure(
            'Shape mismatch', indices,
            lambda np_r, chx_r: (
                'chainerx: {}, numpy: {}'.format(chx_r.shape, np_r.shape)))

    if chainerx_result.device is not chainerx.get_default_device():
        raise _ResultsCheckFailure(
            'ChainerX bad device', indices,
            lambda np_r, chx_r: (
                'default: {}, chainerx: {}'.format(
                    chainerx.get_default_device(), chx_r.device)))

    try:
        check_result_func(chainerx_result, numpy_result)
    except AssertionError as e:
        # Convert AssertionError to _ResultsCheckFailure
        raise _ResultsCheckFailure(str(e), indices)


def _check_chainerx_numpy_result_impl(
        check_result_func, chainerx_result, numpy_result, indices):
    # This function raises _ResultsCheckFailure if any failure occurs.
    # `indices` is a tuple of indices to reach both `chainerx_results` and
    # `numpy_results` from top-level results.

    if chainerx_result is _ignored_result and numpy_result is _ignored_result:
        return

    if isinstance(chainerx_result, (list, tuple)):
        if type(chainerx_result) is not type(numpy_result):
            raise _ResultsCheckFailure('Different result types', indices)
        if len(chainerx_result) != len(numpy_result):
            raise _ResultsCheckFailure('Result length mismatch', indices)
        for i, (chx_r, np_r) in enumerate(zip(chainerx_result, numpy_result)):
            _check_chainerx_numpy_result_impl(
                check_result_func, chx_r, np_r, indices + (i,))

    elif isinstance(chainerx_result, chainerx.ndarray):
        _check_chainerx_numpy_result_array(
            check_result_func, chainerx_result, numpy_result, indices)

    else:
        if _is_numpy_type(chainerx_result):
            raise _ResultsCheckFailure(
                'chainerx result should not be a NumPy type', indices)
        if type(chainerx_result) != type(numpy_result):
            raise _ResultsCheckFailure('Type mismatch', indices)
        if chainerx_result != numpy_result:
            raise _ResultsCheckFailure('Not equal', indices)


def _check_chainerx_numpy_result(
        check_result_func, chainerx_result, numpy_result):
    # Catch _ResultsCheckFailure and generate a comprehensible error message.
    try:
        _check_chainerx_numpy_result_impl(
            check_result_func, chainerx_result, numpy_result, indices=())
    except _ResultsCheckFailure as e:
        indices = e.indices
        chx_r = chainerx_result
        np_r = numpy_result
        i = 0
        while indices[i:]:
            chx_r = chx_r[indices[i]]
            np_r = np_r[indices[i]]
            i += 1

        def make_message(e):
            indices_str = ''.join('[{}]'.format(i) for i in indices)
            s = '{}: {}\n\n'.format(e.msg, e.condense_results(np_r, chx_r))
            if indices:
                s += 'chainerx results{}: {}\n'.format(
                    indices_str, type(chx_r))
                s += '{}\n\n'.format(chx_r)
                s += 'numpy results{}: {}\n'.format(indices_str, type(np_r))
                s += '{}\n\n'.format(np_r)
            s += 'chainerx results: {}\n'.format(type(chainerx_result))
            s += '{}\n\n'.format(chainerx_result)
            s += 'numpy results: {}\n'.format(type(numpy_result))
            s += '{}\n\n'.format(numpy_result)
            return s

        raise AssertionError(make_message(e))


def _make_decorator(check_result_func, name, accept_error):
    def decorator(impl):
        @functools.wraps(impl)
        def test_func(*args, **kw):
            kw[name] = chainerx
            chainerx_result, chainerx_error, chainerx_tb = _call_func(
                impl, args, kw)

            kw[name] = numpy
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', RuntimeWarning)
                numpy_result, numpy_error, numpy_tb = _call_func(
                    impl, args, kw)

            if chainerx_error or numpy_error:
                _check_chainerx_numpy_error(chainerx_error, chainerx_tb,
                                            numpy_error, numpy_tb,
                                            accept_error=accept_error)
                return
            assert chainerx_result is not None and numpy_result is not None, (
                'Either or both of ChainerX and numpy returned None. '
                'chainerx: {}, numpy: {}'.format(
                    chainerx_result, numpy_result))
            _check_chainerx_numpy_result(
                check_result_func, chainerx_result, numpy_result)
        # Apply dummy parametrization on `name` (e.g. 'xp') to avoid pytest
        # error when collecting test functions.
        return pytest.mark.parametrize(name, [None])(test_func)
    return decorator


def numpy_chainerx_allclose(**kwargs):
    """numpy_chainerx_allclose(
           *, rtol=1e-7, atol=0, equal_nan=True, err_msg='', verbose=True,
           name='xp', dtype_check=True, strides_check=True, accept_error=())

    Decorator that checks that NumPy and ChainerX results are equal up to a
    tolerance.

    Args:
         rtol(float): Relative tolerance.
         atol(float): Absolute tolerance.
         equal_nan(bool): Allow NaN values if True. Otherwise, fail the
             assertion if any NaN is found.
         err_msg(str): The error message to be printed in case of failure.
         verbose(bool): If ``True``, the conflicting values are
             appended to the error message.
         name(str): Argument name whose value is either ``numpy`` or
             ``chainerx`` module.
         dtype_check(bool): If ``True``, consistency of dtype is also checked.
             Disabling ``dtype_check`` also implies ``strides_check=False``.
         strides_check(bool): If ``True``, consistency of strides is also
             checked.
         accept_error(Exception or tuple of Exception): Specify
             acceptable errors. When both NumPy test and ChainerX test raises
             the same type of errors, and the type of the errors is specified
             with this argument, the errors are ignored and not raised.
         float16_rtol(float): Relative tolerance for float16 dtype.
         float16_atol(float): Absolute tolerance for float16 dtype.
         float32_rtol(float): Relative tolerance for float32 dtype.
         float32_atol(float): Absolute tolerance for float32 dtype.
         float64_rtol(float): Relative tolerance for float64 dtype.
         float64_atol(float): Absolute tolerance for float64 dtype.

    Decorated test fixture is required to return the same arrays
    in the sense of :func:`numpy_chainerx_allclose`
    (except the type of array module) even if ``xp`` is ``numpy`` or
    ``chainerx``.

    .. seealso:: :func:`chainerx.testing.assert_allclose_ex`
    """  # NOQA
    rtol = kwargs.pop('rtol', 1e-7)
    atol = kwargs.pop('atol', 0)
    equal_nan = kwargs.pop('equal_nan', True)
    err_msg = kwargs.pop('err_msg', '')
    verbose = kwargs.pop('verbose', True)
    name = kwargs.pop('name', 'xp')
    dtype_check = kwargs.pop('dtype_check', None)
    strides_check = kwargs.pop('strides_check', None)
    accept_error = kwargs.pop('accept_error', ())
    tol_kwargs = {k: kwargs[k] for k in
                  ['float16_rtol', 'float32_rtol', 'float64_rtol',
                   'float16_atol', 'float32_atol', 'float64_atol']
                  if k in kwargs}

    def check_result_func(x, y):
        array.assert_allclose_ex(
            x, y, rtol, atol, equal_nan, err_msg, verbose,
            dtype_check=dtype_check, strides_check=strides_check, **tol_kwargs)

    return _make_decorator(check_result_func, name, accept_error)


def numpy_chainerx_array_equal(**kwargs):
    """numpy_chainerx_array_equal(
           *, err_msg='', verbose=True, name='xp', dtype_check=True,
           strides_check=True, accept_error=()):

    Decorator that checks that NumPy and ChainerX results are equal.

    Args:
         err_msg(str): The error message to be printed in case of failure.
         verbose(bool): If ``True``, the conflicting values are
             appended to the error message.
         name(str): Argument name whose value is either ``numpy`` or
             ``chainerx`` module.
         dtype_check(bool): If ``True``, consistency of dtype is also checked.
             Disabling ``dtype_check`` also implies ``strides_check=False``
         strides_check(bool): If ``True``, consistency of strides is also
             checked.
         accept_error(Exception or tuple of Exception): Specify
             acceptable errors. When both NumPy test and ChainerX test raises
             the same type of errors, and the type of the errors is specified
             with this argument, the errors are ignored and not raised.

    Decorated test fixture is required to return the same arrays
    in the sense of :func:`numpy_chainerx_array_equal`
    (except the type of array module) even if ``xp`` is ``numpy`` or
    ``chainerx``.

    .. seealso:: :func:`chainerx.testing.assert_array_equal_ex`
    """
    err_msg = kwargs.pop('err_msg', '')
    verbose = kwargs.pop('verbose', True)
    name = kwargs.pop('name', 'xp')
    dtype_check = kwargs.pop('dtype_check', None)
    strides_check = kwargs.pop('strides_check', None)
    accept_error = kwargs.pop('accept_error', ())

    def check_result_func(x, y):
        array.assert_array_equal_ex(
            x, y, err_msg, verbose, dtype_check=dtype_check,
            strides_check=strides_check)

    return _make_decorator(check_result_func, name, accept_error)



from chainermn.communicators import _memory_utility
from chainermn.communicators import mpi_communicator_base
import numpy as np


class DummyCommunicator(mpi_communicator_base.MpiCommunicatorBase):

    """Dummy communicator that does not communicate at all.

    This class is intended to measure the overhead of packing and unpacking.
    This class does not pass the tests.
    """

    def __init__(self, mpi_comm,
                 batched_copy=False):
        super(DummyCommunicator, self).__init__(mpi_comm)

        self.gpu_buffer_a = _memory_utility.DeviceMemory()
        self.batched_copy = batched_copy

    def multi_node_mean_grad(self, model, zero_fill=False):
        params = _memory_utility.extract_params_set_grad(model, zero_fill)
        itemsize = 4
        n_elems_total = _memory_utility.count_grad_elements(params,
                                                            zero_fill)
        n_bytes_total = n_elems_total * itemsize
        self.gpu_buffer_a.assign(n_bytes_total)

        self._pack_params_to_buffer(params, 'grad',
                                    buffer=self.gpu_buffer_a,
                                    allreduce_grad_dtype=np.float32,
                                    zero_fill=zero_fill)

        self._unpack_params_from_buffer(params, 'grad',
                                        buffer=self.gpu_buffer_a,
                                        allreduce_grad_dtype=np.float32,
                                        zero_fill=zero_fill)
